# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Copernicus-Bench abstract base class."""

import os
import re
from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from typing import Any, Literal

import matplotlib.colors
import numpy as np
import pandas as pd
import rasterio as rio
import torch
from einops import rearrange
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from pyproj import Transformer
from torch import Tensor

from torchgeo.datasets.geo import NonGeoDataset

from ..errors import DatasetNotFoundError, RGBBandsMissingError
from ..utils import (
    Path,
    array_to_tensor,
    disambiguate_timestamp,
    download_and_extract_archive,
    extract_archive,
    percentile_normalization,
)


class CopernicusBenchBase(NonGeoDataset, ABC):
    """Abstract base class for all Copernicus-Bench datasets.

    If you use this dataset in your research, please cite the following paper:

    * https://arxiv.org/abs/2503.11849

    .. versionadded:: 0.7
    """

    @property
    @abstractmethod
    def url(self) -> str:
        """Download URL."""

    #: MD5 checksum.
    md5: str

    #: Zip file name.
    zipfile: str

    #: Subdirectory containing split files.
    directory: str

    #: Filename format of split files.
    filename = '{}.csv'

    #: Mask dtype to cast to, either torch.long for classification
    #: or torch.float for regression.
    dtype: torch.dtype = torch.long

    #: Regular expression used to extract date from filename.
    filename_regex = '.*'

    #: Date format string used to parse date from filename.
    date_format = '%Y%m%dT%H%M%S'

    @property
    @abstractmethod
    def all_bands(self) -> tuple[str, ...]:
        """All spectral channels."""

    @property
    @abstractmethod
    def rgb_bands(self) -> tuple[str, ...]:
        """Spectral channels used to make RGB plots."""

    #: Matplotlib color map for semantic segmentation and change detection plots.
    cmap: str | matplotlib.colors.Colormap

    #: List of classes for classification, semantic segmentation, and change detection.
    classes: tuple[str, ...]

    def __init__(
        self,
        root: Path = 'data',
        split: Literal['train', 'val', 'test'] = 'train',
        bands: Sequence[str] | None = None,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new CopernicusBenchBase instance.

        Args:
            root: Root directory where dataset can be found.
            split: One of 'train', 'val', or 'test'.
            bands: Sequence of band names to load (defaults to all bands).
            transforms: A function/transform that takes input sample and its target as
                entry and returns a transformed version.
            download: If True, download dataset and store it in the root directory.
            checksum: If True, check the MD5 of the downloaded files (may be slow).

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        self.root = root
        self.split = split
        self.bands = bands or self.all_bands
        self.band_indices = [self.all_bands.index(i) + 1 for i in self.bands]
        self.transforms = transforms
        self.download = download
        self.checksum = checksum

        self._verify()

        filepath = os.path.join(root, self.directory, self.filename.format(split))
        self.files = pd.read_csv(filepath, header=None)[0]

    def __len__(self) -> int:
        """Return the length of the dataset.

        Returns:
            Length of the dataset.
        """
        return len(self.files)

    def _load_image(self, path: str) -> dict[str, Tensor]:
        """Load an image and metadata.

        Args:
            path: File path to load.

        Returns:
            An image sample.
        """
        sample: dict[str, Tensor] = {}
        with rio.open(path) as f:
            # Image
            image = f.read(self.band_indices).astype(np.float32)
            sample['image'] = torch.tensor(image)

            # Location
            if f.transform != rio.Affine.identity():
                x = (f.bounds.left + f.bounds.right) / 2
                y = (f.bounds.bottom + f.bounds.top) / 2
                transformer = Transformer.from_crs(f.crs, 'epsg:4326', always_xy=True)
                lon, lat = transformer.transform(x, y)
                sample['lat'] = torch.tensor(lat)
                sample['lon'] = torch.tensor(lon)

            # Time
            if match := re.match(self.filename_regex, os.path.basename(path)):
                if 'date' in match.groupdict():
                    date_str = match.group('date')
                    mint, maxt = disambiguate_timestamp(date_str, self.date_format)
                    time = (mint + maxt) / 2
                    sample['time'] = torch.tensor(time)
                elif 'start' in match.groupdict() and 'stop' in match.groupdict():
                    start = match.group('start')
                    stop = match.group('stop')
                    mint, _ = disambiguate_timestamp(start, self.date_format)
                    _, maxt = disambiguate_timestamp(stop, self.date_format)
                    time = (mint + maxt) / 2
                    sample['time'] = torch.tensor(time)

        return sample

    def _load_mask(self, path: str) -> dict[str, Tensor]:
        """Load a target mask.

        Args:
            path: File path to load.

        Returns:
            A target sample.
        """
        sample: dict[str, Tensor] = {}
        with rio.open(path) as f:
            sample['mask'] = array_to_tensor(f.read(1)).to(self.dtype)

        return sample

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the files already exist
        path = os.path.join(self.root, self.directory, self.filename.format(self.split))
        if os.path.exists(path):
            return

        # Check if the zip file already exists (if so then extract)
        if os.path.exists(os.path.join(self.root, self.zipfile)):
            self._extract()
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download and extract the dataset
        self._download()

    def _extract(self) -> None:
        """Extract the dataset."""
        extract_archive(os.path.join(self.root, self.zipfile))

    def _download(self) -> None:
        """Download the dataset."""
        md5 = self.md5 if self.checksum else None
        download_and_extract_archive(self.url, self.root, md5=md5)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: A sample returned by :meth:`NonGeoDataset.__getitem__`.
            show_titles: Flag indicating whether to show titles above each panel.
            suptitle: Optional string to use as a suptitle.

        Returns:
            A matplotlib Figure with the rendered sample.

        Raises:
            RGBBandsMissingError: If *bands* does not include all RGB bands.
        """
        rgb_indices = []
        for band in self.rgb_bands:
            if band in self.bands:
                rgb_indices.append(self.bands.index(band))
            else:
                raise RGBBandsMissingError()

        # Static -> time series
        images = sample['image'].numpy()
        if sample['image'].dim() == 3:
            images = np.expand_dims(images, axis=0)

        ncols = len(images)
        if 'mask' in sample:
            ncols += 1
            if 'prediction' in sample:
                ncols += 1

        fig, ax = plt.subplots(ncols=ncols, squeeze=False)

        # Label
        title = 'Image'
        if 'label' in sample:
            if sample['label'].dim() == 0:
                # Multiclass classification
                label: Any = self.classes[sample['label']]
                if 'prediction' in sample:
                    prediction: Any = self.classes[sample['prediction']]
            else:
                # Multilabel classification
                label = sample['label'].numpy().nonzero()[0]
                if 'prediction' in sample:
                    prediction = sample['prediction'].numpy().nonzero()[0]

            title = f'Label: {label}'
            if 'prediction' in sample:
                title += f'\nPrediction: {prediction}'

        # Image
        images = images[:, rgb_indices]
        if set(self.rgb_bands) <= {'VV', 'VH', 'HH', 'HV'}:
            # SAR
            vv = images[:, 0]
            vh = images[:, 1]
            images = np.stack([vv, vh, (vv + vh) / 2], axis=1)
            images = percentile_normalization(images)

        images = percentile_normalization(images)
        images = rearrange(images, 't c h w -> t h w c')
        for i in range(len(images)):
            ax[0, i].imshow(images[i])
            ax[0, i].axis('off')
            if show_titles:
                ax[0, i].set_title(title)

        # Mask
        if 'mask' in sample:
            kwargs: dict[str, Any] = {'cmap': self.cmap}
            if hasattr(self, 'classes'):
                # Semantic segmentation
                kwargs |= {
                    'vmin': 0,
                    'vmax': len(self.classes) - 1,
                    'interpolation': 'none',
                }
            mask = sample['mask']
            ax[0, i + 1].imshow(mask, **kwargs)
            ax[0, i + 1].axis('off')
            if show_titles:
                ax[0, i + 1].set_title('Mask')

            if 'prediction' in sample:
                prediction = sample['prediction']
                ax[0, i + 2].imshow(prediction, **kwargs)
                ax[0, i + 2].axis('off')
                if show_titles:
                    ax[0, i + 2].set_title('Prediction')

        if suptitle is not None:
            fig.suptitle(suptitle)

        fig.tight_layout()

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Inria Aerial Image Labeling Dataset."""

import glob
import os
import re
from collections.abc import Callable
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import rasterio as rio
import torch
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, check_integrity, extract_archive, percentile_normalization


class InriaAerialImageLabeling(NonGeoDataset):
    r"""Inria Aerial Image Labeling Dataset.

    The `Inria Aerial Image Labeling <https://project.inria.fr/aerialimagelabeling/>`__
    dataset is a building detection dataset over dissimilar settlements ranging from
    densely populated areas to alpine towns. Refer to the dataset homepage to download
    the dataset.

    Dataset features:

    * Coverage of 810 km\ :sup:`2`\  (405 km\ :sup:`2`\  for training and 405
      km\ :sup:`2`\  for testing)
    * Aerial orthorectified color imagery with a spatial resolution of 0.3 m
    * Number of images: 360 (train: 180, test: 180)
    * Train cities: Austin, Chicago, Kitsap, West Tyrol, Vienna
    * Test cities: Bellingham, Bloomington, Innsbruck, San Francisco, East Tyrol

    Dataset format:

    * Imagery - RGB aerial GeoTIFFs of shape 5000 x 5000
    * Labels - RGB aerial GeoTIFFs of shape 5000 x 5000

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1109/IGARSS.2017.8127684

    .. versionadded:: 0.3

    .. versionchanged:: 0.5
       Added support for a *val* split.
    """

    directory = 'AerialImageDataset'
    filename = 'NEW2-AerialImageDataset.zip'
    md5 = '4b1acfe84ae9961edc1a6049f940380f'

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        checksum: bool = False,
    ) -> None:
        """Initialize a new InriaAerialImageLabeling Dataset instance.

        Args:
            root: root directory where dataset can be found
            split: train/val/test split
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version.
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: if ``split`` is invalid
            DatasetNotFoundError: If dataset is not found.
        """
        self.root = root
        assert split in {'train', 'val', 'test'}
        self.split = split
        self.transforms = transforms
        self.checksum = checksum

        self._verify()
        self.files = self._load_files(root)

    def _load_files(self, root: Path) -> list[dict[str, str]]:
        """Return the paths of the files in the dataset.

        Args:
            root: root dir of dataset

        Returns:
            list of dicts containing paths for each pair of image and label
        """
        files = []
        split = 'train' if self.split in ['train', 'val'] else 'test'
        root_dir = os.path.join(root, self.directory, split)
        pattern = re.compile(r'([A-Za-z]+)(\d+)')

        images = glob.glob(os.path.join(root_dir, 'images', '*.tif'))
        images = sorted(images)

        if split == 'train':
            labels = glob.glob(os.path.join(root_dir, 'gt', '*.tif'))
            labels = sorted(labels)

            for img, lbl in zip(images, labels):
                if match := pattern.search(img):
                    idx = int(match.group(2))
                    # For validation, use the first 5 images of every location
                    if self.split == 'train' and idx > 5:
                        files.append({'image': img, 'label': lbl})
                    elif self.split == 'val' and idx < 6:
                        files.append({'image': img, 'label': lbl})
        else:
            for img in images:
                files.append({'image': img})

        return files

    def _load_image(self, path: Path) -> Tensor:
        """Load a single image.

        Args:
            path: path to the image

        Returns:
            the image
        """
        with rio.open(path) as img:
            array = img.read().astype(np.int32)
            tensor = torch.from_numpy(array).float()
            return tensor

    def _load_target(self, path: Path) -> Tensor:
        """Loads the target mask.

        Args:
            path: path to the mask

        Returns:
            the target mask
        """
        with rio.open(path) as img:
            array = img.read().astype(np.int32)
            array = np.clip(array, 0, 1)
            mask = torch.from_numpy(array[0]).long()
            return mask

    def __len__(self) -> int:
        """Return the number of samples in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.files)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        files = self.files[index]
        img = self._load_image(files['image'])
        sample = {'image': img}
        if files.get('label'):
            mask = self._load_target(files['label'])
            sample['mask'] = mask

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _verify(self) -> None:
        """Checks the integrity of the dataset structure."""
        if os.path.isdir(os.path.join(self.root, self.directory)):
            return

        archive_path = os.path.join(self.root, self.filename)
        md5_hash = self.md5 if self.checksum else None
        if not os.path.isfile(archive_path):
            raise DatasetNotFoundError(self)
        if not check_integrity(archive_path, md5_hash):
            raise RuntimeError('Dataset corrupted')
        print('Extracting...')
        extract_archive(archive_path)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        image = np.rollaxis(sample['image'][:3].numpy(), 0, 3)
        image = percentile_normalization(image, axis=(0, 1))

        ncols = 1
        show_mask = 'mask' in sample
        show_predictions = 'prediction' in sample

        if show_mask:
            mask = sample['mask'].numpy()
            ncols += 1

        if show_predictions:
            prediction = sample['prediction'].numpy()
            ncols += 1

        fig, axs = plt.subplots(ncols=ncols, figsize=(ncols * 8, 8))
        if not isinstance(axs, np.ndarray):
            axs = [axs]
        axs[0].imshow(image)
        axs[0].axis('off')
        if show_titles:
            axs[0].set_title('Image')

        if show_mask:
            axs[1].imshow(mask, interpolation='none')
            axs[1].axis('off')
            if show_titles:
                axs[1].set_title('Label')

        if show_predictions:
            axs[2].imshow(prediction, interpolation='none')
            axs[2].axis('off')
            if show_titles:
                axs[2].set_title('Prediction')

        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""MMFlood dataset."""

from __future__ import annotations

import os
from collections.abc import Callable
from glob import glob
from typing import ClassVar, Literal

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.figure import Figure
from rasterio.crs import CRS
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import IntersectionDataset, RasterDataset
from .utils import BoundingBox, Path, download_url, extract_archive


class MMFloodComponent(RasterDataset):
    """Base component for MMFlood dataset."""

    def __init__(
        self,
        subfolders: list[str],
        content: Literal['s1_raw', 'DEM', 'hydro', 'mask'],
        root: Path = 'data',
        crs: CRS | None = None,
        res: float | tuple[float, float] | None = None,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        cache: bool = False,
    ) -> None:
        """Initialize MMFloodComponent dataset instance.

        Args:
            subfolders: list of directories to be loaded
            content: specifies which component to load
            root: root directory where dataset can be found
            crs: :term:`coordinate reference system (CRS)` to warp to
                (defaults to the CRS of the first file found)
            res: resolution of the dataset in units of CRS in (xres, yres) format. If a
                single float is provided, it is used for both the x and y resolution.
                (defaults to the resolution of the first file found)
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            cache: if True, cache file handle to speed up repeated sampling
        """
        self.content = content
        self.is_image = content != 'mask'
        paths = []
        for s in subfolders:
            paths += glob(os.path.join(root, '**', f'{s}*-*', self.content, '*.tif'))
        paths = sorted(paths)
        super().__init__(paths, crs, res, transforms=transforms, cache=cache)


class MMFloodIntersection(IntersectionDataset):
    """Intersection dataset used to merge two or more MMFloodComponents."""

    def __init__(
        self,
        dataset1: MMFloodComponent | MMFloodIntersection,
        dataset2: MMFloodComponent | MMFloodIntersection,
    ) -> None:
        """Initialize a new MMFloodIntersection instance.

        Args:
            dataset1: the first dataset to merge
            dataset2: the second dataset to merge
        """
        # if hydro component is passed, it should always be passed as dataset2
        super().__init__(dataset1, dataset2)

    def _merge_dataset_indices(self) -> None:
        """Create a new R-tree out of the individual indices from Sentinel-1, DEM and hydrography datasets."""
        _, ds2 = self.datasets
        # Always use index of ds2, since it either coincides with ds1 index
        # or refers to hydro, which represents only a subset of the dataset
        self.index = ds2.index


class MMFlood(IntersectionDataset):
    """MMFlood dataset.

    `MMFlood <https://huggingface.co/datasets/links-ads/mmflood>`__ dataset is a multimodal
    flood delineation dataset. Sentinel-1 data is matched with masks and DEM data for all
    available tiles. If hydrography maps are loaded, only a subset of the dataset is loaded,
    since only 1,012 Sentinel-1 tiles have a corresponding hydrography map.
    Some Sentinel-1 tiles have missing data, which are automatically set to 0.
    Corresponding pixels in masks are set to 255 and should be ignored in performance computation.

    Dataset features:

    * 1,748 Sentinel-1 tiles of varying pixel dimensions
    * multimodal dataset
    * 95 flood events from 42 different countries
    * includes DEMs
    * includes hydrography maps (available for 1,012 tiles out of 1,748)
    * flood delineation maps (ground truth) is obtained from Copernicus EMS

    Dataset classes:

    0. no flood
    1. flood

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1109/ACCESS.2022.3205419

    .. versionadded:: 0.7
    """

    url = 'https://huggingface.co/datasets/links-ads/mmflood/resolve/24ca097306c9e50ad0711903c11e1ba13ea1bedc/'
    _ignore_index = 255
    _nparts = 11

    metadata: ClassVar[dict[str, str]] = {
        'part_file': 'activations.tar.{part}.gz.part',
        'filename': 'activations.tar.gz',
        'directory': 'activations',
        'metadata_file': 'activations.json',
    }
    _splits: ClassVar[set[str]] = {'train', 'val', 'test'}
    _md5: ClassVar[dict[str, str]] = {
        'activations.json': 'de33a3ac7e55a0051ada21cbdfbb4745',
        'activations.tar.gz': '3cd4c4fe7506aa40263f74639d85ccce',
        'activations.tar.000.gz.part': 'a8424653edca6e79999831bdda53d4dc',
        'activations.tar.001.gz.part': '517def8760d3ce86885c7600c77a1d6c',
        'activations.tar.002.gz.part': '6797b97121f5b98ff58fde7491f584b2',
        'activations.tar.003.gz.part': 'e69d2a6b1746ef869d1da4d22018a71a',
        'activations.tar.004.gz.part': '0ccf7ea69ea6c0e88db1b1015ec3361e',
        'activations.tar.005.gz.part': '8ef6765afe20f254b1e752d7a2742fda',
        'activations.tar.006.gz.part': '3f330a44b66511b7a95f4a555f8b793a',
        'activations.tar.007.gz.part': '1d2046b5f3c473c3681a05dc94b29b86',
        'activations.tar.008.gz.part': 'f386b5acf78f8ae34592404c6c7ec43c',
        'activations.tar.009.gz.part': 'dd5317a3c0d33de815beadb9850baa38',
        'activations.tar.010.gz.part': '5a14a7e3f916c5dcf288c2ca88daf4d0',
    }

    def __init__(
        self,
        root: Path = 'data',
        crs: CRS | None = None,
        res: float | tuple[float, float] | None = None,
        split: str = 'train',
        include_dem: bool = False,
        include_hydro: bool = False,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
        cache: bool = False,
    ) -> None:
        """Initialize a new MMFlood dataset instance.

        Args:
            root: root directory where dataset can be found
            crs: :term:`coordinate reference system (CRS)` to warp to
                (defaults to the CRS of the first file found)
            res: resolution of the dataset in units of CRS in (xres, yres) format. If a
                single float is provided, it is used for both the x and y resolution.
                (defaults to the resolution of the first file found)
            split: train/val/test split to load
            include_dem: If True, DEM data is concatenated after Sentinel-1 bands.
            include_hydro: If True, hydrography data is concatenated as last channel.
                Only a smaller subset of the original dataset is loaded in this case.
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)
            cache: if True, cache file handle to speed up repeated sampling

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
            AssertionError: If *split* is invalid.

        """
        assert split in self._splits

        self.root = root
        self.split = split
        self.include_dem = include_dem
        self.include_hydro = include_hydro
        self.transforms = transforms
        self.download = download
        self.checksum = checksum
        # Verify integrity of the dataset
        self._verify()
        self.metadata_df = pd.read_json(
            os.path.join(self.root, self.metadata['metadata_file'])
        ).transpose()

        split_subfolders = self.metadata_df[
            self.metadata_df['subset'] == self.split
        ].index.tolist()
        self.image: MMFloodComponent | MMFloodIntersection = MMFloodComponent(
            split_subfolders, 's1_raw', root, crs, res, cache=cache
        )
        if include_dem:
            dem = MMFloodComponent(split_subfolders, 'DEM', root, crs, res, cache=cache)
            self.image = MMFloodIntersection(self.image, dem)
        if include_hydro:
            hydro = MMFloodComponent(
                split_subfolders, 'hydro', root, crs, res, cache=cache
            )
            self.image = MMFloodIntersection(self.image, hydro)
        self.mask = MMFloodComponent(
            split_subfolders, 'mask', root, crs, res, cache=cache
        )

        super().__init__(self.image, self.mask, transforms=transforms)

    def _merge_tar_files(self) -> None:
        """Merge part tar gz files."""
        dst_filename = self.metadata['filename']
        dst_path = os.path.join(self.root, dst_filename)

        print('Merging separate part files...')
        with open(dst_path, 'wb') as dst_fp:
            for idx in range(self._nparts):
                part_filename = f'activations.tar.{idx:03}.gz.part'
                part_path = os.path.join(self.root, part_filename)
                print(f'Processing file {part_path!s}')

                with open(part_path, 'rb') as part_fp:
                    dst_fp.write(part_fp.read())

    def __getitem__(self, query: BoundingBox) -> dict[str, Tensor]:
        """Retrieve image/mask and metadata indexed by query.

        Args:
            query: (minx, maxx, miny, maxy, mint, maxt) coordinates to index

        Returns:
            sample of image, mask and metadata at that index

        Raises:
            IndexError: if query is not found in the index
        """
        data = super().__getitem__(query)
        missing_data = data['image'].isnan().any(dim=0)
        # Set all pixel values of invalid areas to 0, all mask values to 255
        data['image'][:, missing_data] = 0
        data['mask'][missing_data] = self._ignore_index
        return data

    def _merge_dataset_indices(self) -> None:
        """Create a new R-tree out of the individual indices from Sentinel-1, DEM and hydrography datasets."""
        ds1, _ = self.datasets
        # Use ds1 index
        self.index = ds1.index

    def _download(self) -> None:
        """Download the dataset."""

        def _check_and_download(filename: str, url: str) -> None:
            path = os.path.join(self.root, filename)
            if not os.path.exists(path):
                md5 = self._md5[filename] if self.checksum else None
                download_url(url, self.root, filename, md5)
            return

        filename = self.metadata['filename']
        filepath = os.path.join(self.root, filename)
        if not os.path.exists(filepath):
            for idx in range(self._nparts):
                part_file = f'activations.tar.{idx:03}.gz.part'
                url = self.url + part_file

                _check_and_download(part_file, url)

        _check_and_download(
            self.metadata['metadata_file'], self.url + self.metadata['metadata_file']
        )

    def _extract(self) -> None:
        """Extract the dataset."""
        filepath = os.path.join(self.root, self.metadata['filename'])
        if str(filepath).endswith('.tar.gz'):
            extract_archive(filepath)

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        dirpath = os.path.join(self.root, self.metadata['directory'])
        metadata_filepath = os.path.join(self.root, self.metadata['metadata_file'])
        # Check if both metadata file and directory exist
        if os.path.isdir(dirpath) and os.path.isfile(metadata_filepath):
            return
        if not self.download:
            raise DatasetNotFoundError(self)
        self._download()
        self._merge_tar_files()
        self._extract()

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample
        """
        show_mask = 'mask' in sample
        image = sample['image'][[0, 1]].permute(1, 2, 0).numpy()
        ncols = 1
        show_predictions = 'prediction' in sample
        if self.include_dem:
            dem_idx = -2 if self.include_hydro else -1
            dem = sample['image'][dem_idx].squeeze(0).numpy()
            ncols += 1
        if self.include_hydro:
            hydro = sample['image'][-1].squeeze(0).numpy()
            ncols += 1
        if show_mask:
            mask = sample['mask'].numpy()
            # Set ignore_index values to 0
            mask[mask == self._ignore_index] = 0
            ncols += 1
        if show_predictions:
            pred = sample['prediction'].numpy()
            ncols += 1

        # Compute False Color image, from Sentinel1 plot function
        co_polarization = image[..., 0]  # transmit == receive
        cross_polarization = image[..., 1]  # transmit != receive
        ratio = co_polarization / cross_polarization

        # https://gis.stackexchange.com/a/400780/123758
        co_polarization = np.clip(co_polarization / 0.3, a_min=0, a_max=1)
        cross_polarization = np.clip(cross_polarization / 0.05, a_min=0, a_max=1)
        ratio = np.clip(ratio / 25, a_min=0, a_max=1)

        image = np.stack((co_polarization, cross_polarization, ratio), axis=-1)

        # Generate the figure
        fig, axs = plt.subplots(ncols=ncols, figsize=(4 * ncols, 4))
        axs[0].imshow(image)
        axs[0].axis('off')
        axs_idx = 1
        if self.include_dem:
            axs[axs_idx].imshow(dem, cmap='gray')
            axs[axs_idx].axis('off')
            axs_idx += 1
        if self.include_hydro:
            axs[axs_idx].imshow(hydro, cmap='gray')
            axs[axs_idx].axis('off')
            axs_idx += 1
        if show_mask:
            axs[axs_idx].imshow(mask, cmap='gray')
            axs[axs_idx].axis('off')
            axs_idx += 1
        if show_predictions:
            axs[axs_idx].imshow(pred, cmap='gray')
            axs[axs_idx].axis('off')

        if show_titles:
            axs[0].set_title('Image')
            axs_idx = 1
            if self.include_dem:
                axs[axs_idx].set_title('DEM')
                axs_idx += 1
            if self.include_hydro:
                axs[axs_idx].set_title('Hydrography Map')
                axs_idx += 1
            if show_mask:
                axs[axs_idx].set_title('Mask')
                axs_idx += 1
            if show_predictions:
                axs[axs_idx].set_title('Prediction')

        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""TorchGeo batch samplers."""

import abc
from collections.abc import Iterator

import torch
from rtree.index import Index, Property
from torch import Generator
from torch.utils.data import Sampler

from ..datasets import BoundingBox, GeoDataset
from .constants import Units
from .utils import _to_tuple, get_random_bounding_box, tile_to_chips


class BatchGeoSampler(Sampler[list[BoundingBox]], abc.ABC):
    """Abstract base class for sampling from :class:`~torchgeo.datasets.GeoDataset`.

    Unlike PyTorch's :class:`~torch.utils.data.BatchSampler`, :class:`BatchGeoSampler`
    returns enough geospatial information to uniquely index any
    :class:`~torchgeo.datasets.GeoDataset`. This includes things like latitude,
    longitude, height, width, projection, coordinate system, and time.
    """

    def __init__(self, dataset: GeoDataset, roi: BoundingBox | None = None) -> None:
        """Initialize a new Sampler instance.

        Args:
            dataset: dataset to index from
            roi: region of interest to sample from (minx, maxx, miny, maxy, mint, maxt)
                (defaults to the bounds of ``dataset.index``)
        """
        if roi is None:
            self.index = dataset.index
            roi = BoundingBox(*self.index.bounds)
        else:
            self.index = Index(interleaved=False, properties=Property(dimension=3))
            hits = dataset.index.intersection(tuple(roi), objects=True)
            for hit in hits:
                bbox = BoundingBox(*hit.bounds) & roi
                self.index.insert(hit.id, tuple(bbox), hit.object)

        self.res = dataset.res
        self.roi = roi

    @abc.abstractmethod
    def __iter__(self) -> Iterator[list[BoundingBox]]:
        """Return a batch of indices of a dataset.

        Returns:
            batch of (minx, maxx, miny, maxy, mint, maxt) coordinates to index a dataset
        """


class RandomBatchGeoSampler(BatchGeoSampler):
    """Samples batches of elements from a region of interest randomly.

    This is particularly useful during training when you want to maximize the size of
    the dataset and return as many random :term:`chips <chip>` as possible. Note that
    randomly sampled chips may overlap.
    """

    def __init__(
        self,
        dataset: GeoDataset,
        size: tuple[float, float] | float,
        batch_size: int,
        length: int | None = None,
        roi: BoundingBox | None = None,
        units: Units = Units.PIXELS,
        generator: Generator | None = None,
    ) -> None:
        """Initialize a new Sampler instance.

        The ``size`` argument can either be:

        * a single ``float`` - in which case the same value is used for the height and
          width dimension
        * a ``tuple`` of two floats - in which case, the first *float* is used for the
          height dimension, and the second *float* for the width dimension

        .. versionchanged:: 0.3
           Added ``units`` parameter, changed default to pixel units

        .. versionchanged:: 0.4
           ``length`` parameter is now optional, a reasonable default will be used

        .. versionadded:: 0.7
            The *generator* parameter.

        Args:
            dataset: dataset to index from
            size: dimensions of each :term:`patch`
            batch_size: number of samples per batch
            length: number of samples per epoch
                (defaults to approximately the maximal number of non-overlapping
                :term:`chips <chip>` of size ``size`` that could be sampled from
                the dataset)
            roi: region of interest to sample from (minx, maxx, miny, maxy, mint, maxt)
                (defaults to the bounds of ``dataset.index``)
            units: defines if ``size`` is in pixel or CRS units
            generator: pseudo-random number generator (PRNG).
        """
        super().__init__(dataset, roi)
        self.size = _to_tuple(size)
        self.generator = generator

        if units == Units.PIXELS:
            self.size = (self.size[0] * self.res[1], self.size[1] * self.res[0])

        self.batch_size = batch_size
        self.length = 0
        self.hits = []
        areas = []
        for hit in self.index.intersection(tuple(self.roi), objects=True):
            bounds = BoundingBox(*hit.bounds)
            if (
                bounds.maxx - bounds.minx >= self.size[1]
                and bounds.maxy - bounds.miny >= self.size[0]
            ):
                if bounds.area > 0:
                    rows, cols = tile_to_chips(bounds, self.size)
                    self.length += rows * cols
                else:
                    self.length += 1
                self.hits.append(hit)
                areas.append(bounds.area)
        if length is not None:
            self.length = length

        # torch.multinomial requires float probabilities > 0
        self.areas = torch.tensor(areas, dtype=torch.float)
        if torch.sum(self.areas) == 0:
            self.areas += 1

    def __iter__(self) -> Iterator[list[BoundingBox]]:
        """Return the indices of a dataset.

        Returns:
            batch of (minx, maxx, miny, maxy, mint, maxt) coordinates to index a dataset
        """
        for _ in range(len(self)):
            # Choose a random tile, weighted by area
            idx = torch.multinomial(self.areas, 1)
            hit = self.hits[idx]
            bounds = BoundingBox(*hit.bounds)

            # Choose random indices within that tile
            batch = []
            for _ in range(self.batch_size):
                bounding_box = get_random_bounding_box(
                    bounds, self.size, self.res, self.generator
                )
                batch.append(bounding_box)

            yield batch

    def __len__(self) -> int:
        """Return the number of batches in a single epoch.

        Returns:
            number of batches in an epoch
        """
        return self.length // self.batch_size



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Copernicus-Bench BigEarthNet-S1 dataset."""

import os
from collections.abc import Callable, Sequence
from typing import Literal

import numpy as np
import pandas as pd
import torch
from torch import Tensor

from ..utils import Path
from .base import CopernicusBenchBase


class CopernicusBenchBigEarthNetS1(CopernicusBenchBase):
    """Copernicus-Bench BigEarthNet-S1 dataset.

    BigEarthNet-S1 is a multilabel land use/land cover classification dataset
    composed of 5% of the Sentinel-1 data of BigEarthNet-v2.

    If you use this dataset in your research, please cite the following papers:

    * https://arxiv.org/abs/2503.11849
    * https://arxiv.org/abs/2407.03653

    .. versionadded:: 0.7
    """

    url = 'https://hf.co/datasets/wangyi111/Copernicus-Bench/resolve/9d252acd3aa0e3da3128e05c6f028647f0e48e5f/l2_bigearthnet_s1s2/bigearthnetv2.zip'
    md5 = '269355db0449e0da7213c95f30c346d4'
    zipfile = 'bigearthnetv2.zip'
    directory = 'bigearthnet_s1s2'
    filename = 'multilabel-{}.csv'
    filename_regex = r'.{16}_(?P<date>\d{8}T\d{6})'
    all_bands = ('VV', 'VH')
    rgb_bands = ('VV', 'VH')
    classes = (
        'Urban fabric',
        'Industrial or commercial units',
        'Arable land',
        'Permanent crops',
        'Pastures',
        'Complex cultivation patterns',
        'Land principally occupied by agriculture, with significant areas of natural vegetation',
        'Agro-forestry areas',
        'Broad-leaved forest',
        'Coniferous forest',
        'Mixed forest',
        'Natural grassland and sparsely vegetated areas',
        'Moors, heathland and sclerophyllous vegetation',
        'Transitional woodland, shrub',
        'Beaches, dunes, sands',
        'Inland wetlands',
        'Coastal wetlands',
        'Inland waters',
        'Marine waters',
    )

    def __init__(
        self,
        root: Path = 'data',
        split: Literal['train', 'val', 'test'] = 'train',
        bands: Sequence[str] | None = None,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new CopernicusBenchBigEarthNetS1 instance.

        Args:
            root: Root directory where dataset can be found.
            split: One of 'train', 'val', or 'test'.
            bands: Sequence of band names to load (defaults to all bands).
            transforms: A function/transform that takes input sample and its target as
                entry and returns a transformed version.
            download: If True, download dataset and store it in the root directory.
            checksum: If True, check the MD5 of the downloaded files (may be slow).

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        super().__init__(root, split, bands, transforms, download, checksum)
        filepath = os.path.join(root, self.directory, self.filename.format(split))
        self.files = pd.read_csv(filepath)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: Index to return.

        Returns:
            Data and labels at that index.
        """
        row = self.files.iloc[index].values
        file = row[0]
        path = os.path.join(self.root, self.directory, 'BigEarthNet-S1-5%', file)
        sample = self._load_image(path)
        sample['label'] = torch.tensor(row[2:].astype(np.int64))

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Million-AID dataset."""

import glob
import os
from collections.abc import Callable
from typing import Any, ClassVar, cast

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.figure import Figure
from PIL import Image
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, check_integrity, extract_archive


class MillionAID(NonGeoDataset):
    """Million-AID Dataset.

    The `MillionAID <https://captain-whu.github.io/DiRS/>`_ dataset consists
    of one million aerial images from Google Earth Engine that offers
    either `a multi-class learning task
    <https://competitions.codalab.org/competitions/35945#learn_the_details-dataset>`_
    with 51 classes or a `multi-label learning task
    <https://competitions.codalab.org/competitions/35974#learn_the_details-dataset>`_
    with 73 different possible labels. For more details please consult
    the accompanying `paper <https://ieeexplore.ieee.org/document/9393553>`_.

    Dataset features:

    * RGB aerial images with varying resolutions from 0.5 m to 153 m per pixel
    * images within classes can have different pixel dimension

    Dataset format:

    * images are three-channel jpg

    If you use this dataset in your research, please cite the following paper:

    * https://ieeexplore.ieee.org/document/9393553

    .. versionadded:: 0.3
    """

    multi_label_categories = (
        'agriculture_land',
        'airport_area',
        'apartment',
        'apron',
        'arable_land',
        'bare_land',
        'baseball_field',
        'basketball_court',
        'beach',
        'bridge',
        'cemetery',
        'church',
        'commercial_area',
        'commercial_land',
        'dam',
        'desert',
        'detached_house',
        'dry_field',
        'factory_area',
        'forest',
        'golf_course',
        'grassland',
        'greenhouse',
        'ground_track_field',
        'helipad',
        'highway_area',
        'ice_land',
        'industrial_land',
        'intersection',
        'island',
        'lake',
        'leisure_land',
        'meadow',
        'mine',
        'mining_area',
        'mobile_home_park',
        'oil_field',
        'orchard',
        'paddy_field',
        'parking_lot',
        'pier',
        'port_area',
        'power_station',
        'public_service_land',
        'quarry',
        'railway',
        'railway_area',
        'religious_land',
        'residential_land',
        'river',
        'road',
        'rock_land',
        'roundabout',
        'runway',
        'solar_power_plant',
        'sparse_shrub_land',
        'special_land',
        'sports_land',
        'stadium',
        'storage_tank',
        'substation',
        'swimming_pool',
        'tennis_court',
        'terraced_field',
        'train_station',
        'transportation_land',
        'unutilized_land',
        'viaduct',
        'wastewater_plant',
        'water_area',
        'wind_turbine',
        'woodland',
        'works',
    )

    multi_class_categories = (
        'apartment',
        'apron',
        'bare_land',
        'baseball_field',
        'bapsketball_court',
        'beach',
        'bridge',
        'cemetery',
        'church',
        'commercial_area',
        'dam',
        'desert',
        'detached_house',
        'dry_field',
        'forest',
        'golf_course',
        'greenhouse',
        'ground_track_field',
        'helipad',
        'ice_land',
        'intersection',
        'island',
        'lake',
        'meadow',
        'mine',
        'mobile_home_park',
        'oil_field',
        'orchard',
        'paddy_field',
        'parking_lot',
        'pier',
        'quarry',
        'railway',
        'river',
        'road',
        'rock_land',
        'roundabout',
        'runway',
        'solar_power_plant',
        'sparse_shrub_land',
        'stadium',
        'storage_tank',
        'substation',
        'swimming_pool',
        'tennis_court',
        'terraced_field',
        'train_station',
        'viaduct',
        'wastewater_plant',
        'wind_turbine',
        'works',
    )

    md5s: ClassVar[dict[str, str]] = {
        'train': '1b40503cafa9b0601653ca36cd788852',
        'test': '51a63ee3eeb1351889eacff349a983d8',
    }

    filenames: ClassVar[dict[str, str]] = {'train': 'train.zip', 'test': 'test.zip'}

    tasks = ('multi-class', 'multi-label')
    splits = ('train', 'test')

    def __init__(
        self,
        root: Path = 'data',
        task: str = 'multi-class',
        split: str = 'train',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        checksum: bool = False,
    ) -> None:
        """Initialize a new MillionAID dataset instance.

        Args:
            root: root directory where dataset can be found
            task: type of task, either "multi-class" or "multi-label"
            split: train or test split
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found.
        """
        self.root = root
        self.transforms = transforms
        self.checksum = checksum
        assert task in self.tasks
        assert split in self.splits
        self.task = task
        self.split = split

        self._verify()

        self.files = self._load_files(self.root)

        self.classes = sorted({cls for f in self.files for cls in f['label']})
        self.class_to_idx: dict[str, int] = {c: i for i, c in enumerate(self.classes)}

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.files)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        files = self.files[index]
        image = self._load_image(files['image'])
        cls_label = [self.class_to_idx[label] for label in files['label']]
        label = torch.tensor(cls_label, dtype=torch.long)
        sample = {'image': image, 'label': label}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _load_files(self, root: Path) -> list[dict[str, Any]]:
        """Return the paths of the files in the dataset.

        Args:
            root: root directory of dataset

        Returns:
            list of dicts containing paths for each pair of image, and list of labels
        """
        imgs_no_subcat = list(
            glob.glob(os.path.join(root, self.split, '*', '*', '*.jpg'))
        )

        imgs_subcat = list(
            glob.glob(os.path.join(root, self.split, '*', '*', '*', '*.jpg'))
        )

        scenes = [p.split(os.sep)[-3] for p in imgs_no_subcat] + [
            p.split(os.sep)[-4] for p in imgs_subcat
        ]

        subcategories = ['Missing' for p in imgs_no_subcat] + [
            p.split(os.sep)[-3] for p in imgs_subcat
        ]

        classes = [p.split(os.sep)[-2] for p in imgs_no_subcat] + [
            p.split(os.sep)[-2] for p in imgs_subcat
        ]

        if self.task == 'multi-label':
            labels = [
                [sc, sub, c] if sub != 'Missing' else [sc, c]
                for sc, sub, c in zip(scenes, subcategories, classes)
            ]
        else:
            labels = [[c] for c in classes]

        images = imgs_no_subcat + imgs_subcat

        files = [dict(image=image, label=label) for image, label in zip(images, labels)]

        return files

    def _load_image(self, path: Path) -> Tensor:
        """Load a single image.

        Args:
            path: path to the image

        Returns:
            the image
        """
        with Image.open(path) as img:
            array: np.typing.NDArray[np.int_] = np.array(img.convert('RGB'))
            tensor: Tensor = torch.from_numpy(array)
            # Convert from HxWxC to CxHxW
            tensor = tensor.permute((2, 0, 1))
            return tensor

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        filepath = os.path.join(self.root, self.split)
        if os.path.isdir(filepath):
            return

        filepath = os.path.join(self.root, self.split + '.zip')
        if os.path.isfile(filepath):
            if self.checksum and not check_integrity(filepath, self.md5s[self.split]):
                raise RuntimeError('Dataset found, but corrupted.')
            extract_archive(filepath)
            return

        raise DatasetNotFoundError(self)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample

        """
        image = np.rollaxis(sample['image'].numpy(), 0, 3)
        labels = [self.classes[cast(int, label)] for label in sample['label']]

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            prediction_labels = [
                self.classes[cast(int, label)] for label in sample['prediction']
            ]

        fig, ax = plt.subplots(figsize=(4, 4))
        ax.imshow(image)
        ax.axis('off')
        if show_titles:
            title = f'Label: {labels}'
            if showing_predictions:
                title += f'\nPrediction: {prediction_labels}'
            ax.set_title(title)

        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""GlobBiomass dataset."""

import glob
import os
from collections.abc import Callable, Iterable
from typing import Any, ClassVar, cast

import matplotlib.pyplot as plt
import torch
from matplotlib.figure import Figure
from rasterio.crs import CRS

from .errors import DatasetNotFoundError
from .geo import RasterDataset
from .utils import (
    BoundingBox,
    Path,
    check_integrity,
    disambiguate_timestamp,
    extract_archive,
)


class GlobBiomass(RasterDataset):
    """GlobBiomass dataset.

    The `GlobBiomass <https://doi.pangaea.de/10.1594/PANGAEA.894711>`__ dataset consists
    of global pixelwise aboveground biomass (AGB) and growth stock volume (GSV) maps.

    Definitions:

    * AGB: the mass, expressed as oven-dry weight of the woody parts
      (stem, bark, branches and twigs) of all living trees excluding stump and roots.
    * GSV: volume of all living trees more than 10 cm in diameter at breast height
      measured over bark from ground or stump height to a top stem diameter of 0 cm.

    Units:

    * AGB: m3/ha
    * GSV: tons/ha (i.e., Mg/ha)

    Dataset features:

    * Global estimates of AGB and GSV at ~100 m per pixel resolution
      (45,000 x 45,000 px)
    * Per-pixel uncertainty expressed as standard error

    Dataset format:

    * Estimate maps are single-channel
    * Uncertainty maps are single-channel

    The data can be manually downloaded from `this website
    <https://globbiomass.org/wp-content/uploads/GB_Maps/Globbiomass_global_dataset.html>`_.

    If you use this dataset in your research, please cite the following dataset:

    * https://doi.org/10.1594/PANGAEA.894711

    .. versionadded:: 0.3
    """

    filename_glob = '*_{}.tif'
    filename_regex = r"""
        ^(?P<tile>[NS][\d]{2}[EW][\d]{3})
        _(?P<measurement>(agb|gsv))
    """
    mint, maxt = disambiguate_timestamp('2010', '%Y')
    is_image = False
    dtype = torch.float32  # pixelwise regression

    measurements = ('agb', 'gsv')

    md5s: ClassVar[dict[str, str]] = {
        'N00E020_agb.zip': 'bd83a3a4c143885d1962bde549413be6',
        'N00E020_gsv.zip': 'da5ddb88e369df2d781a0c6be008ae79',
        'N00E060_agb.zip': '85eaca95b939086cc528e396b75bd097',
        'N00E060_gsv.zip': 'ec84174697c17ca4db2967374446ab30',
        'N00E100_agb.zip': 'c50c7c996615c1c6f19cb383ef11812a',
        'N00E100_gsv.zip': '6e0ff834db822d3710ed40d00a200e8f',
        'N00E140_agb.zip': '73f0b44b9e137789cefb711ef9aa281b',
        'N00E140_gsv.zip': '43be3dd4563b63d12de006d240ba5edf',
        'N00W020_agb.zip': '4fb979732f0a22cc7a2ca3667698084b',
        'N00W020_gsv.zip': 'ac5bbeedaa0f94a5e01c7a86751d6891',
        'N00W060_agb.zip': '59da0b32b08fbbcd2dd76926a849562b',
        'N00W060_gsv.zip': '5ca9598f621a7d10ab1d623ee5b44aa6',
        'N00W100_agb.zip': 'a819b75a39e8d4d37b15745c96ea1e35',
        'N00W100_gsv.zip': '71aad3669d522f7190029ec33350831a',
        'N00W180_agb.zip': '5a1d7486d8310fbaf4980a76e9ffcd78',
        'N00W180_gsv.zip': '274be7dbb4e6d7563773cc302129a9c7',
        'N40E020_agb.zip': '38bc7170f94734b365d614a566f872e7',
        'N40E020_gsv.zip': 'b52c1c777d68c331cc058a273530536e',
        'N40E060_agb.zip': '1d94ad59f3f26664fefa4d7308b63f05',
        'N40E060_gsv.zip': '3b68786b7641400077ef340a7ef748f4',
        'N40E100_agb.zip': '3ccb436047c0db416fb237435645989c',
        'N40E100_gsv.zip': 'c44efe9e7ce2ae0f2e39b0db10f06c71',
        'N40E140_agb.zip': '35ea51da229af1312ba4aaafc0dbd5d6',
        'N40E140_gsv.zip': '8431828708c84263a4971a8779864f69',
        'N40W020_agb.zip': '38345a1826719301ab1a0251b4835cc2',
        'N40W020_gsv.zip': '5e136b7c2f921cd425cb5cc5669e7693',
        'N40W060_agb.zip': 'e3f54df1d188c0132ecf5aef3dc54ca6',
        'N40W060_gsv.zip': '09093d78ffef0220cb459a88e61e3093',
        'N40W100_agb.zip': 'cc21ce8793e5594dc7a0b45f0d0f1466',
        'N40W100_gsv.zip': '21be1398df88818d04dcce422e2010a6',
        'N40W140_agb.zip': '64665f53fad7386abb1cf4a44a1c8b1a',
        'N40W140_gsv.zip': 'b59405219fc807cbe745789fbb6936a6',
        'N40W180_agb.zip': 'f83ef786da8333ee739e49be108994c1',
        'N40W180_gsv.zip': '1f2eb8912b1a204eaeb2858b7e398baa',
        'N80E020_agb.zip': '7f7aed44802890672bd908e28eda6f15',
        'N80E020_gsv.zip': '6e285eec66306e56dc3a81adc0da2a27',
        'N80E060_agb.zip': '55e7031e0207888f25f27efa9a0ab8f4',
        'N80E060_gsv.zip': '8d14c7f61ad2aed527e124f9aacae30c',
        'N80E100_agb.zip': '562eafd2813ff06e47284c48324bb1c7',
        'N80E100_gsv.zip': '73067e0fac442c330ae2294996280042',
        'N80E140_agb.zip': '1b51ce0df0dba925c5ef2149bebca115',
        'N80E140_gsv.zip': '37ee3047d281fc34fa3a9e024a8317a1',
        'N80W020_agb.zip': '60dde6adc0dfa219a34c976367f571c0',
        'N80W020_gsv.zip': 'b7be4e97bb4179710291ee8dee27f538',
        'N80W060_agb.zip': 'db7d35d0375851c4a181c3a8fa8b480e',
        'N80W060_gsv.zip': 'd36ffcf4622348382454c979baf53234',
        'N80W100_agb.zip': 'c0dbf53e635dabf9a4d7d1756adeda69',
        'N80W100_gsv.zip': 'abdeaf0d65da1216c326b6d0ce27d61b',
        'N80W140_agb.zip': '7719c0efd23cd86215fea0285fd0ea4a',
        'N80W140_gsv.zip': '499969bed381197ee9427a2e3f455a2e',
        'N80W180_agb.zip': 'e3a163d1944e1989a07225d262a01c6f',
        'N80W180_gsv.zip': '5d39ec0368cfe63c40c66d61ae07f577',
        'S40E140_agb.zip': '263eb077a984117b41cc7cfa0c32915b',
        'S40E140_gsv.zip': 'e0ffad85fbade4fb711cc5b3c7543898',
        'S40W060_agb.zip': '2cbf6858c48f36add896db660826829b',
        'S40W060_gsv.zip': '04dbfd4aca0bd2a2a7d8f563c8659252',
        'S40W100_agb.zip': 'ae89f021e7d9c2afea433878f77d1dd6',
        'S40W100_gsv.zip': 'b6aa3f276e1b51dade803a71df2acde6',
    }

    def __init__(
        self,
        paths: Path | Iterable[Path] = 'data',
        crs: CRS | None = None,
        res: float | tuple[float, float] | None = None,
        measurement: str = 'agb',
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        cache: bool = True,
        checksum: bool = False,
    ) -> None:
        """Initialize a new GlobBiomass instance.

        Args:
            paths: one or more root directories to search or files to load
            crs: :term:`coordinate reference system (CRS)` to warp to
                (defaults to the CRS of the first file found)
            res: resolution of the dataset in units of CRS in (xres, yres) format. If a
                single float is provided, it is used for both the x and y resolution.
                (defaults to the resolution of the first file found)
            measurement: use data from 'agb' or 'gsv' measurement
            transforms: a function/transform that takes an input sample
                and returns a transformed version
            cache: if True, cache file handle to speed up repeated sampling
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: If *measurement* is not valid.
            DatasetNotFoundError: If dataset is not found.

        .. versionchanged:: 0.5
           *root* was renamed to *paths*.
        """
        assert measurement in self.measurements

        self.paths = paths
        self.measurement = measurement
        self.checksum = checksum

        self.filename_glob = self.filename_glob.format(measurement)

        self._verify()

        super().__init__(paths, crs, res, transforms=transforms, cache=cache)

    def __getitem__(self, query: BoundingBox) -> dict[str, Any]:
        """Retrieve image/mask and metadata indexed by query.

        Args:
            query: (minx, maxx, miny, maxy, mint, maxt) coordinates to index

        Returns:
            sample at index consisting of measurement mask with 2 channels,
            where the first is the measurement and the second the error map

        Raises:
            IndexError: if query is not found in the index
        """
        hits = self.index.intersection(tuple(query), objects=True)
        filepaths = cast(list[str], [hit.object for hit in hits])

        if not filepaths:
            raise IndexError(
                f'query: {query} not found in index with bounds: {self.bounds}'
            )

        mask = self._merge_files(filepaths, query)

        std_error_paths = [str(f).replace('.tif', '_err.tif') for f in filepaths]
        std_err_mask = self._merge_files(std_error_paths, query)

        mask = torch.cat((mask, std_err_mask), dim=0)

        sample = {'mask': mask, 'crs': self.crs, 'bounds': query}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted file already exists
        if self.files:
            return

        # Check if the zip files have already been downloaded
        assert isinstance(self.paths, str | os.PathLike)
        pathname = os.path.join(self.paths, f'*_{self.measurement}.zip')
        if glob.glob(pathname):
            for zipfile in glob.iglob(pathname):
                filename = os.path.basename(zipfile)
                if self.checksum and not check_integrity(zipfile, self.md5s[filename]):
                    raise RuntimeError('Dataset found, but corrupted.')
                extract_archive(zipfile)
            return

        raise DatasetNotFoundError(self)

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        tensor = sample['mask']
        mask = tensor[0, ...]
        error_mask = tensor[1, ...]

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            pred = sample['prediction'][0, ...]
            ncols = 3
        else:
            ncols = 2

        fig, axs = plt.subplots(nrows=1, ncols=ncols, figsize=(ncols * 4, 4))

        if showing_predictions:
            axs[0].imshow(mask)
            axs[0].axis('off')
            axs[1].imshow(error_mask)
            axs[1].axis('off')
            axs[2].imshow(pred)
            axs[2].axis('off')
            if show_titles:
                axs[0].set_title('Mask')
                axs[1].set_title('Uncertainty Mask')
                axs[2].set_title('Prediction')
        else:
            axs[0].imshow(mask)
            axs[0].axis('off')
            axs[1].imshow(error_mask)
            axs[1].axis('off')
            if show_titles:
                axs[0].set_title('Mask')
                axs[1].set_title('Uncertainty Mask')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""SpaceNet 5 dataset."""

from typing import ClassVar

from .spacenet3 import SpaceNet3


class SpaceNet5(SpaceNet3):
    r"""SpaceNet 5: Automated Road Network Extraction and Route Travel Time Estimation.

    `SpaceNet 5 <https://spacenet.ai/sn5-challenge/>`_
    is a dataset of road networks over the cities of Moscow, Mumbai and San
    Juan (unavailable).

    Collection features:

    +------------+---------------------+------------+---------------------------+
    |    AOI     | Area (km\ :sup:`2`\)| # Images   | # Road Network Labels (km)|
    +============+=====================+============+===========================+
    | Moscow     |    1353             |   1353     |         3066              |
    +------------+---------------------+------------+---------------------------+
    | Mumbai     |    1021             |   1016     |         1951              |
    +------------+---------------------+------------+---------------------------+

    Imagery features:

    .. list-table::
        :widths: 10 10 10 10 10
        :header-rows: 1
        :stub-columns: 1

        *   -
            - PAN
            - MS
            - PS-MS
            - PS-RGB
        *   - GSD (m)
            - 0.31
            - 1.24
            - 0.30
            - 0.30
        *   - Chip size (px)
            - 1300 x 1300
            - 325 x 325
            - 1300 x 1300
            - 1300 x 1300

    Dataset format:

    * Imagery - Worldview-3 GeoTIFFs

        * PAN.tif (Panchromatic)
        * MS.tif (Multispectral)
        * PS-MS (Pansharpened Multispectral)
        * PS-RGB (Pansharpened RGB)

    * Labels - GeoJSON

        * labels.geojson

    If you use this dataset in your research, please use the following citation:

    * The SpaceNet Partners, “SpaceNet5: Automated Road Network Extraction and
      Route Travel Time Estimation from Satellite Imagery”,
      https://spacenet.ai/sn5-challenge/

    .. versionadded:: 0.2
    """

    file_regex = r'_chip(\d+)\.'
    dataset_id = 'SN5_roads'
    tarballs: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {
            7: ['SN5_roads_train_AOI_7_Moscow.tar.gz'],
            8: ['SN5_roads_train_AOI_8_Mumbai.tar.gz'],
        },
        'test': {9: ['SN5_roads_test_public_AOI_9_San_Juan.tar.gz']},
    }
    md5s: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {
            7: ['03082d01081a6d8df2bc5a9645148d2a'],
            8: ['1ee20ba781da6cb7696eef9a95a5bdcc'],
        },
        'test': {9: ['fc45afef219dfd3a20f2d4fc597f6882']},
    }
    valid_aois: ClassVar[dict[str, list[int]]] = {'train': [7, 8], 'test': [9]}
    valid_images: ClassVar[dict[str, list[str]]] = {
        'train': ['MS', 'PAN', 'PS-MS', 'PS-RGB'],
        'test': ['MS', 'PAN', 'PS-MS', 'PS-RGB'],
    }
    valid_masks = ('geojson_roads_speed',)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Rwanda Field Boundary Competition dataset."""

import glob
import os
from collections.abc import Callable, Sequence
from typing import ClassVar

import matplotlib.pyplot as plt
import numpy as np
import rasterio
import rasterio.features
import torch
from einops import rearrange
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError, RGBBandsMissingError
from .geo import NonGeoDataset
from .utils import Path, which


class RwandaFieldBoundary(NonGeoDataset):
    """Rwanda Field Boundary Competition dataset.

    This dataset contains field boundaries for smallholder farms in eastern Rwanda.
    The Nasa Harvest program funded a team of annotators from TaQadam to label Planet
    imagery for the 2021 growing season for the purpose of conducting the Rwanda Field
    boundary detection Challenge. The dataset includes rasterized labeled field
    boundaries and time series satellite imagery from Planet's NICFI program.
    Planet's basemap imagery is provided for six months (March, April, August, October,
    November and December). Note: only fields that were big enough to be differentiated
    on the Planetscope imagery were labeled, only fields that were fully contained
    within the chips were labeled. The paired dataset is provided in 256x256 chips for a
    total of 70 tiles covering 1532 individual fields.

    The labels are provided as binary semantic segmentation labels:

    0. No field-boundary
    1. Field-boundary

    If you use this dataset in your research, please cite the following:

    * https://doi.org/10.34911/RDNT.G580WW

    .. note::

       This dataset requires the following additional library to be installed:

       * `azcopy <https://github.com/Azure/azure-storage-azcopy>`_: to download the
         dataset from Source Cooperative.

    .. versionadded:: 0.5
    """

    url = 'https://radiantearth.blob.core.windows.net/mlhub/nasa_rwanda_field_boundary_competition'

    splits: ClassVar[dict[str, int]] = {'train': 57, 'test': 13}
    dates = ('2021_03', '2021_04', '2021_08', '2021_10', '2021_11', '2021_12')
    all_bands = ('B01', 'B02', 'B03', 'B04')
    rgb_bands = ('B03', 'B02', 'B01')
    classes = ('No field-boundary', 'Field-boundary')

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        bands: Sequence[str] = all_bands,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
    ) -> None:
        """Initialize a new RwandaFieldBoundary instance.

        Args:
            root: root directory where dataset can be found
            split: one of "train" or "test"
            bands: the subset of bands to load
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory

        Raises:
            AssertionError: If *split* or *bands* are invalid.
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert split in self.splits
        assert set(bands) <= set(self.all_bands)

        self.root = root
        self.split = split
        self.bands = bands
        self.transforms = transforms
        self.download = download

        self._verify()

    def __len__(self) -> int:
        """Return the number of chips in the dataset.

        Returns:
            length of the dataset
        """
        return self.splits[self.split]

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            a dict containing image and mask at index.
        """
        images = []
        for date in self.dates:
            patches = []
            for band in self.bands:
                path = os.path.join(self.root, 'source', self.split, date)
                with rasterio.open(os.path.join(path, f'{index:02}_{band}.tif')) as src:
                    patches.append(src.read(1).astype(np.float32))
            images.append(patches)
        sample = {'image': torch.from_numpy(np.array(images))}

        if self.split == 'train':
            path = os.path.join(self.root, 'labels', self.split)
            with rasterio.open(os.path.join(path, f'{index:02}.tif')) as src:
                sample['mask'] = torch.from_numpy(src.read(1).astype(np.int64))

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the subdirectories already exist and have the correct number of files
        path = os.path.join(self.root, 'source', self.split, '*', '*.tif')
        expected = len(self.dates) * self.splits[self.split] * len(self.all_bands)
        if len(glob.glob(path)) == expected:
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download and extract the dataset
        self._download()

    def _download(self) -> None:
        """Download the dataset."""
        os.makedirs(self.root, exist_ok=True)
        azcopy = which('azcopy')
        azcopy('sync', self.url, self.root, '--recursive=true')

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        time_step: int = 0,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            time_step: time step at which to access image, beginning with 0
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample

        Raises:
            RGBBandsMissingError: If *bands* does not include all RGB bands.
        """
        rgb_indices = []
        for band in self.rgb_bands:
            if band in self.bands:
                rgb_indices.append(self.bands.index(band))
            else:
                raise RGBBandsMissingError()

        ncols = 1
        for key in ('mask', 'prediction'):
            if key in sample:
                ncols += 1

        fig, axs = plt.subplots(ncols=ncols, squeeze=False)

        image = torch.clamp(sample['image'][time_step, rgb_indices] / 2000, 0, 1)
        image = rearrange(image, 'c h w -> h w c')
        axs[0, 0].imshow(image)
        axs[0, 0].axis('off')
        if show_titles:
            axs[0, 0].set_title(f't={time_step}')

        if 'mask' in sample:
            axs[0, 1].imshow(sample['mask'])
            axs[0, 1].axis('off')
            if show_titles:
                axs[0, 1].set_title('Mask')

        if 'prediction' in sample:
            axs[0, 2].imshow(sample['prediction'])
            axs[0, 2].axis('off')
            if show_titles:
                axs[0, 2].set_title('Prediction')

        if suptitle is not None:
            fig.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Airphen dataset."""

from typing import Any

import matplotlib.pyplot as plt
from matplotlib.figure import Figure

from .errors import RGBBandsMissingError
from .geo import RasterDataset
from .utils import percentile_normalization


class Airphen(RasterDataset):
    """Airphen dataset.

    `Airphen <https://ideol.sakura.ne.jp/img/20170123_HiphenAirphenKeyfeatures.pdf>`__
    is a multispectral scientific camera developed by agronomists and photonics
    engineers at `Hiphen <https://www.hiphen-plant.com/>`_ to match plant measurements
    needs and constraints.

    Main characteristics:

    * 6 Synchronized global shutter sensors
    * Sensor resolution 1280 x 960 pixels
    * Data format (.tiff, 12 bit)
    * SD card storage
    * Metadata information: Exif and XMP
    * Internal or external GPS
    * Synchronization with different sensors (TIR, RGB, others)

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.34133/2021/9892647

    .. versionadded:: 0.6
    """

    # Each camera measures a custom set of spectral bands chosen at purchase time.
    # Hiphen offers 8 bands to choose from, sorted from short to long wavelength.
    all_bands = ('B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', 'B8')
    rgb_bands = ('B4', 'B3', 'B1')

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`RasterDataset.__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample

        Raises:
            RGBBandsMissingError: If *bands* does not include all RGB bands.
        """
        rgb_indices = []
        for band in self.rgb_bands:
            if band in self.bands:
                rgb_indices.append(self.bands.index(band))
            else:
                raise RGBBandsMissingError()

        image = sample['image'][rgb_indices].permute(1, 2, 0).float()
        image = percentile_normalization(image, axis=(0, 1))

        fig, ax = plt.subplots(1, 1, figsize=(4, 4))
        ax.imshow(image)
        ax.axis('off')

        if show_titles:
            ax.set_title('Image')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Seasonal Contrast datamodule."""

from typing import Any

import kornia.augmentation as K
import torch
from einops import repeat

from ..datasets import SeasonalContrastS2
from .geo import NonGeoDataModule


class SeasonalContrastS2DataModule(NonGeoDataModule):
    """LightningDataModule implementation for the Seasonal Contrast dataset.

    .. versionadded:: 0.5
    """

    def __init__(
        self, batch_size: int = 64, num_workers: int = 0, **kwargs: Any
    ) -> None:
        """Initialize a new SeasonalContrastS2DataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.SeasonalContrastS2`.
        """
        super().__init__(SeasonalContrastS2, batch_size, num_workers, **kwargs)

        bands = kwargs.get('bands', SeasonalContrastS2.rgb_bands)
        seasons = kwargs.get('seasons', 1)

        # Normalization only available for RGB dataset, defined here:
        # https://github.com/ServiceNow/seasonal-contrast/blob/8285173ec205b64bc3e53b880344dd6c3f79fa7a/datasets/seco_dataset.py
        if bands == SeasonalContrastS2.rgb_bands:
            _min = torch.tensor([3, 2, 0])
            _max = torch.tensor([88, 103, 129])
            _mean = torch.tensor([0.485, 0.456, 0.406])
            _std = torch.tensor([0.229, 0.224, 0.225])

            _min = repeat(_min, 'c -> (t c)', t=seasons)
            _max = repeat(_max, 'c -> (t c)', t=seasons)
            _mean = repeat(_mean, 'c -> (t c)', t=seasons)
            _std = repeat(_std, 'c -> (t c)', t=seasons)

            self.aug = K.AugmentationSequential(
                K.Normalize(mean=_min, std=_max - _min),
                K.Normalize(mean=torch.tensor(0), std=1 / torch.tensor(255)),
                K.Normalize(mean=_mean, std=_std),
                data_keys=None,
                keepdim=True,
            )

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        self.dataset = SeasonalContrastS2(**self.kwargs)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""CaFFe datamodule."""

from typing import Any

import kornia.augmentation as K
import torch

from ..datasets import CaFFe
from .geo import NonGeoDataModule


class CaFFeDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the CaFFe dataset.

    Implements the default splits that come with the dataset.

    .. versionadded:: 0.7
    """

    mean = torch.Tensor([0.5517])
    std = torch.Tensor([11.8478])

    def __init__(
        self, batch_size: int = 64, num_workers: int = 0, size: int = 512, **kwargs: Any
    ) -> None:
        """Initialize a new CaFFeDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            size: resize images of input size 512x512 to size x size
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.CaFFe`.
        """
        super().__init__(CaFFe, batch_size, num_workers, **kwargs)

        self.size = size

        self.train_aug = K.AugmentationSequential(
            K.Normalize(mean=self.mean, std=self.std),
            K.Resize(size),
            K.RandomHorizontalFlip(p=0.5),
            K.RandomVerticalFlip(p=0.5),
            data_keys=None,
            keepdim=True,
        )

        self.aug = K.AugmentationSequential(
            K.Normalize(mean=self.mean, std=self.std),
            K.Resize(size),
            data_keys=None,
            keepdim=True,
        )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Pre-trained Scale-MAE models."""

from collections import OrderedDict
from functools import partial
from typing import Any

import kornia.augmentation as K
import torch
import torch.nn as nn
from timm.models.vision_transformer import VisionTransformer
from torch import Tensor
from torchvision.models._api import Weights, WeightsEnum

_mean = torch.tensor([0.485, 0.456, 0.406])
_std = torch.tensor([0.229, 0.224, 0.225])
_scale_mae_transforms = K.AugmentationSequential(
    K.Normalize(mean=torch.tensor(0), std=torch.tensor(255)),
    K.Normalize(mean=_mean, std=_std),
    data_keys=None,
)


def get_2d_sincos_pos_embed_with_resolution(
    embed_dim: int, grid_size: int, res: Tensor, cls_token: bool = False
) -> Tensor:
    """Generate spatial resolution specific 2D positional embeddings.

    Args:
        embed_dim: Dimension of the positional embeddings.
        grid_size: Height (ph) and width (pw) of the image patches.
        res: Spatial resolution tensor of shape (N,) of the image.
        cls_token: Increase positional embedding size by 1 for class token.

    Returns:
        pos_embed: Spatial resolution aware positional embeddings (Ph * Pw, D).
    """
    device, dtype = res.device, res.dtype
    grid_h = torch.arange(grid_size, dtype=dtype, device=device)
    grid_w = torch.arange(grid_size, dtype=dtype, device=device)
    grid: Tensor = torch.stack(torch.meshgrid(grid_w, grid_h, indexing='xy'), dim=0)
    grid = torch.einsum('chw,n->cnhw', grid, res)
    _, n, h, w = grid.shape
    pos_embed = get_2d_sincos_pos_embed_from_grid_torch(embed_dim, grid)
    pos_embed = pos_embed.reshape(n, h * w, embed_dim)
    if cls_token:
        pos_embed = torch.cat(
            [torch.zeros([n, 1, embed_dim], dtype=dtype, device=device), pos_embed],
            dim=1,
        )
    return pos_embed


def get_2d_sincos_pos_embed_from_grid_torch(embed_dim: int, grid: Tensor) -> Tensor:
    """Generate 2D sin-cos positional embedding from grid.

    Args:
        embed_dim: Dimension of the positional embeddings.
        grid: Tensor representing the image patch grid (C, N, Ph, Pw)

    Returns:
        emb: 2D sin-cos positional embeddings (Ph * Pw, D).
    """
    assert embed_dim % 2 == 0
    emb_h = get_1d_sincos_pos_embed_from_grid_torch(embed_dim // 2, grid[0])
    emb_w = get_1d_sincos_pos_embed_from_grid_torch(embed_dim // 2, grid[1])
    emb = torch.cat([emb_h, emb_w], dim=1)
    return emb


def get_1d_sincos_pos_embed_from_grid_torch(embed_dim: int, pos: Tensor) -> Tensor:
    """Generate 1D sin-cos positional embedding from grid dimension.

    Args:
        embed_dim: Dimension of the positional embeddings.
        pos: Tensor of positions to be encoded (M,).

    Returns:
        emb: 1D sin-cos positional embeddings (M, D).
    """
    assert embed_dim % 2 == 0
    omega = torch.arange(embed_dim // 2, dtype=pos.dtype, device=pos.device)
    omega /= embed_dim / 2.0
    omega = 1.0 / 10000**omega
    pos = pos.reshape(-1)
    out = torch.einsum('m,d->md', pos, omega)
    emb_sin = torch.sin(out)
    emb_cos = torch.cos(out)
    emb = torch.cat([emb_sin, emb_cos], dim=1)
    return emb


class ScaleMAE(VisionTransformer):
    """Custom Vision Transformer for Scale-MAE with GSD positional embeddings.

    This is a ViT encoder only model of the Scale-MAE architecture with GSD positional embeddings.

    If you use this model in your research, please cite the following paper:

    * https://arxiv.org/abs/2212.14532
    """

    def __init__(self, res: float = 1.0, *args: Any, **kwargs: Any) -> None:
        """Initialize a new ScaleMAE model.

        Args:
            res: Spatial resolution of the image in meters.
            *args: Additional arguments to
                pass to :class:`timm.models.vision_transformer.VisionTransformer`.
            **kwargs: Additional keyword arguments to
                pass to :class:`timm.models.vision_transformer.VisionTransformer`.
        """
        super().__init__(*args, **kwargs)

        self.res = res

        # Scale MAE uses resolution specific positional embeddings
        if self.pos_embed is not None:
            self.pos_embed.requires_grad = False

    def _pos_embed(self, x: Tensor) -> Tensor:
        """Apply GSD positional embeddings to the input tensor."""
        res = torch.tensor(self.res, dtype=x.dtype, device=x.device)
        res = res.repeat(x.shape[0])
        pos_embed = (
            get_2d_sincos_pos_embed_with_resolution(
                self.embed_dim,
                int(self.patch_embed.num_patches**0.5),
                res,
                cls_token=True,
            )
            .to(x.dtype)
            .to(x.device)
        )
        if self.cls_token is not None:
            cls_tokens = self.cls_token.expand(x.shape[0], -1, -1)
            x = torch.cat((cls_tokens, x), dim=1)
        x = x + pos_embed
        x = self.pos_drop(x)
        return x


def interpolate_pos_embed(
    model: ScaleMAE, state_dict: OrderedDict[str, Tensor]
) -> OrderedDict[str, Tensor]:
    """Interpolate the positional embeddings if image size is different than pretrained image size.

    Args:
        model: ScaleMAE model.
        state_dict: Pretrained model state dict.

    Returns:
        state_dict: State dict with interpolated positional embeddings.
    """
    pos_embed_checkpoint = state_dict['pos_embed']
    embedding_size = pos_embed_checkpoint.shape[-1]
    num_patches = model.patch_embed.num_patches
    num_extra_tokens = 0
    if model.pos_embed is not None:
        num_extra_tokens = model.pos_embed.shape[-2] - num_patches
    # height (== width) for the checkpoint position embedding
    orig_size = int((pos_embed_checkpoint.shape[-2] - num_extra_tokens) ** 0.5)
    # height (== width) for the new position embedding
    new_size = int(num_patches**0.5)
    # class_token and dist_token are kept unchanged
    if orig_size != new_size:
        print(
            f'Interpolating positional embeddings from {orig_size}x{orig_size} to {new_size}x{new_size}'
        )
        extra_tokens = pos_embed_checkpoint[:, :num_extra_tokens]
        # only the position tokens are interpolated
        pos_tokens = pos_embed_checkpoint[:, num_extra_tokens:]
        pos_tokens = pos_tokens.reshape(
            -1, orig_size, orig_size, embedding_size
        ).permute(0, 3, 1, 2)
        pos_tokens = torch.nn.functional.interpolate(
            pos_tokens, size=(new_size, new_size), mode='bicubic', align_corners=False
        )
        pos_tokens = pos_tokens.permute(0, 2, 3, 1).flatten(1, 2)
        new_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=1)
        state_dict['pos_embed'] = new_pos_embed

    return state_dict


# https://github.com/pytorch/vision/pull/6883
# https://github.com/pytorch/vision/pull/7107
# Can be removed once torchvision>=0.15 is required
Weights.__deepcopy__ = lambda *args, **kwargs: args[0]


class ScaleMAELarge16_Weights(WeightsEnum):  # type: ignore[misc]
    """Scale-MAE Large patch size 16 weights.

    .. versionadded:: 0.6
    """

    FMOW_RGB = Weights(
        url='https://hf.co/torchgeo/vit_large_patch16_224_fmow_rgb_scalemae/resolve/9dc7f569424baeb780698352cf6e87638c882123/vit_large_patch16_224_fmow_rgb_scalemae-98ed9821.pth',
        transforms=_scale_mae_transforms,
        meta={
            'dataset': 'fMoW',
            'in_chans': 3,
            'img_size': 224,
            'model': 'vit_large_patch16',
            'publication': 'https://arxiv.org/abs/2212.14532',
            'repo': 'https://github.com/bair-climate-initiative/scale-mae',
        },
    )


def scalemae_large_patch16(
    weights: ScaleMAELarge16_Weights | None = None, *args: Any, **kwargs: Any
) -> ScaleMAE:
    """Scale-MAE Large model.

    If you use this model in your research, please cite the following paper:

    * https://arxiv.org/abs/2212.14532

    .. versionadded:: 0.6

    Args:
        weights: Pre-trained model weights to use.
        *args: Additional arguments to
            pass to :class:`ScaleMAE`.
        **kwargs: Additional keyword arguments to
            pass to :class:`ScaleMAE`.

    Returns:
        A Scale-MAE Large patch16 model.
    """
    model = ScaleMAE(
        patch_size=16,
        embed_dim=1024,
        depth=24,
        num_heads=16,
        mlp_ratio=4,
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        *args,
        **kwargs,
    )

    if weights:
        state_dict = weights.get_state_dict(progress=True)

        if 'img_size' in kwargs and weights.meta['img_size'] != kwargs['img_size']:
            state_dict = interpolate_pos_embed(model, state_dict)

        model.load_state_dict(state_dict, strict=False)

    return model



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Copernicus-Bench LC100Seg-S3 dataset."""

import glob
import os
from collections.abc import Callable, Sequence
from typing import ClassVar, Literal

import pandas as pd
from matplotlib.colors import ListedColormap
from torch import Tensor

from ..utils import Path, stack_samples
from .base import CopernicusBenchBase


class CopernicusBenchLC100SegS3(CopernicusBenchBase):
    """Copernicus-Bench LC100Seg-S3 dataset.

    LC100Seg-S3 is a multilabel land use/land cover segmentation dataset based on
    Sentinel-3 OLCI images and CGLS-LC100 land cover maps. CGLS-LC100 is a product in
    the Copernicus Global Land Service (CGLS) portfolio and delivers a global 23-class
    land cover map at 100 m spatial resolution.

    This benchmark supports both static (1 image/location) and time series
    (1-4 images/location) modes, the former is used in the original benchmark.

    .. list-table:: Classes
       :header-rows: 1

       * - Value
         - Description
       * - 0
         - Unknown. No or not enough satellite data available.
       * - 20
         - Shrubs. Woody perennial plants with persistent and woody stems and without
           any defined main stem being less than 5 m tall. The shrub foliage can be
           either evergreen or deciduous.
       * - 30
         - Herbaceous vegetation. Plants without persistent stem or shoots above ground
           and lacking definite firm structure. Tree and shrub cover is less than 10 %.
       * - 40
         - Cultivated and managed vegetation / agriculture. Lands covered with temporary
           crops followed by harvest and a bare soil period (e.g., single and multiple
           cropping systems). Note that perennial woody crops will be classified as the
           appropriate forest or shrub land cover type.
       * - 50
         - Urban / built up. Land covered by buildings and other man-made structures.
       * - 60
         - Bare / sparse vegetation. Lands with exposed soil, sand, or rocks and never
           has more than 10 % vegetated cover during any time of the year.
       * - 70
         - Snow and ice. Lands under snow or ice cover throughout the year.
       * - 80
         - Permanent water bodies. Lakes, reservoirs, and rivers. Can be either fresh
           or salt-water bodies.
       * - 90
         - Herbaceous wetland. Lands with a permanent mixture of water and herbaceous
           or woody vegetation. The vegetation can be present in either salt, brackish,
           or fresh water.
       * - 100
         - Moss and lichen.
       * - 111
         - Closed forest, evergreen needle leaf. Tree canopy >70 %, almost all needle
           leaf trees remain green all year. Canopy is never without green foliage.
       * - 112
         - Closed forest, evergreen broad leaf. Tree canopy >70 %, almost all broadleaf
           trees remain green year round. Canopy is never without green foliage.
       * - 113
         - Closed forest, deciduous needle leaf. Tree canopy >70 %, consists of seasonal
           needle leaf tree communities with an annual cycle of leaf-on and leaf-off
           periods.
       * - 114
         - Closed forest, deciduous broad leaf. Tree canopy >70 %, consists of seasonal
           broadleaf tree communities with an annual cycle of leaf-on and leaf-off
           periods.
       * - 115
         - Closed forest, mixed.
       * - 116
         - Closed forest, not matching any of the other definitions.
       * - 121
         - Open forest, evergreen needle leaf. Top layer- trees 15-70 % and second
           layer- mixed of shrubs and grassland, almost all needle leaf trees remain
           green all year. Canopy is never without green foliage.
       * - 122
         - Open forest, evergreen broad leaf. Top layer- trees 15-70 % and second layer-
           mixed of shrubs and grassland, almost all broadleaf trees remain green year
           round. Canopy is never without green foliage.
       * - 123
         - Open forest, deciduous needle leaf. Top layer- trees 15-70 % and second
           layer- mixed of shrubs and grassland, consists of seasonal needle leaf tree
           communities with an annual cycle of leaf-on and leaf-off periods.
       * - 124
         - Open forest, deciduous broad leaf. Top layer- trees 15-70 % and second layer-
           mixed of shrubs and grassland, consists of seasonal broadleaf tree
           communities with an annual cycle of leaf-on and leaf-off periods.
       * - 125
         - Open forest, mixed.
       * - 126
         - Open forest, not matching any of the other definitions.
       * - 200
         - Oceans, seas. Can be either fresh or salt-water bodies.

    If you use this dataset in your research, please cite the following papers:

    * https://arxiv.org/abs/2503.11849
    * https://doi.org/10.5281/zenodo.3939049

    .. versionadded:: 0.7
    """

    url = 'https://hf.co/datasets/wangyi111/Copernicus-Bench/resolve/9d252acd3aa0e3da3128e05c6f028647f0e48e5f/l2_lc100_s3/lc100_s3.zip'
    md5 = '967d1da6286e0d0e346e425a8f3800e9'
    zipfile = 'lc100_s3.zip'
    filename = 'multilabel-{}.csv'
    directory = 'lc100_s3'
    filename_regex = r'S3[AB]_(?P<date>\d{8}T\d{6})'
    all_bands = (
        'Oa01_radiance',
        'Oa02_radiance',
        'Oa03_radiance',
        'Oa04_radiance',
        'Oa05_radiance',
        'Oa06_radiance',
        'Oa07_radiance',
        'Oa08_radiance',
        'Oa09_radiance',
        'Oa10_radiance',
        'Oa11_radiance',
        'Oa12_radiance',
        'Oa13_radiance',
        'Oa14_radiance',
        'Oa15_radiance',
        'Oa16_radiance',
        'Oa17_radiance',
        'Oa18_radiance',
        'Oa19_radiance',
        'Oa20_radiance',
        'Oa21_radiance',
    )
    rgb_bands = ('Oa08_radiance', 'Oa06_radiance', 'Oa04_radiance')
    cmap = ListedColormap(
        [
            '#282828',
            '#ffbb22',
            '#ffff4c',
            '#f096ff',
            '#fa0000',
            '#b4b4b4',
            '#f0f0f0',
            '#0032c8',
            '#0096a0',
            '#fae6a0',
            '#58481f',
            '#009900',
            '#70663e',
            '#00cc00',
            '#4e751f',
            '#007800',
            '#666000',
            '#8db400',
            '#8d7400',
            '#a0dc00',
            '#929900',
            '#648c00',
            '#000080',
        ]
    )
    classes = (
        'Unknown',
        'Shrubs',
        'Herbaceous vegetation',
        'Cultivated and managed vegetation / agriculture',
        'Urban / built up',
        'Bare / sparse vegetation',
        'Snow and ice',
        'Permanent water bodies',
        'Herbaceous wetland',
        'Moss and lichen',
        'Closed forest, evergreen needle leaf',
        'Closed forest, evergreen broad leaf',
        'Closed forest, deciduous needle leaf',
        'Closed forest, deciduous broad leaf',
        'Closed forest, mixed',
        'Closed forest, not matching any of the other definitions',
        'Open forest, evergreen needle leaf',
        'Open forest, evergreen broad leaf',
        'Open forest, deciduous needle leaf',
        'Open forest, deciduous broad leaf',
        'Open forest, mixed',
        'Open forest, not matching any of the other definitions',
        'Oceans, seas',
    )
    class_map: ClassVar[dict[int, int]] = {
        0: 0,
        20: 1,
        30: 2,
        40: 3,
        50: 4,
        60: 5,
        70: 6,
        80: 7,
        90: 8,
        100: 9,
        111: 10,
        112: 11,
        113: 12,
        114: 13,
        115: 14,
        116: 15,
        121: 16,
        122: 17,
        123: 18,
        124: 19,
        125: 20,
        126: 21,
        200: 22,
    }

    def __init__(
        self,
        root: Path = 'data',
        split: Literal['train', 'val', 'test'] = 'train',
        mode: Literal['static', 'time-series'] = 'static',
        bands: Sequence[str] | None = None,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new CopernicusBenchLC100SegS3 instance.

        Args:
            root: Root directory where dataset can be found.
            split: One of 'train', 'val', or 'test'.
            mode: One of 'static' or 'time-series'.
            bands: Sequence of band names to load (defaults to all bands).
            transforms: A function/transform that takes input sample and its target as
                entry and returns a transformed version.
            download: If True, download dataset and store it in the root directory.
            checksum: If True, check the MD5 of the downloaded files (may be slow).

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        self.mode = mode
        super().__init__(root, split, bands, transforms, download, checksum)
        filepath = os.path.join(root, self.directory, self.filename.format(split))
        self.files = pd.read_csv(filepath)
        if mode == 'static':
            filepath = os.path.join(root, self.directory, f'static_fnames-{split}.csv')
            self.static_files = pd.read_csv(filepath, header=None)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: Index to return.

        Returns:
            Data and labels at that index.
        """
        row = self.files.iloc[index].values
        match self.mode:
            case 'static':
                pid, file = self.static_files.iloc[index]
                path = os.path.join(self.root, self.directory, 's3_olci', pid, file)
                sample = self._load_image(path)
            case 'time-series':
                pid = row[0]
                paths = os.path.join(self.root, self.directory, 's3_olci', pid, '*.tif')
                samples = [self._load_image(path) for path in sorted(glob.glob(paths))]
                sample = stack_samples(samples)

        path = os.path.join(self.root, self.directory, 'lc100', f'{pid}.tif')
        sample |= self._load_mask(path)
        for before, after in self.class_map.items():
            sample['mask'][sample['mask'] == before] = after

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Trainers for regression."""

import os
from typing import Any

import kornia.augmentation as K
import matplotlib.pyplot as plt
import segmentation_models_pytorch as smp
import timm
import torch
import torch.nn as nn
from matplotlib.figure import Figure
from torch import Tensor
from torchmetrics import MeanAbsoluteError, MeanSquaredError, MetricCollection
from torchvision.models._api import WeightsEnum

from ..datasets import RGBBandsMissingError, unbind_samples
from ..models import FCN, get_weight
from . import utils
from .base import BaseTask


class RegressionTask(BaseTask):
    """Regression."""

    target_key = 'label'

    def __init__(
        self,
        model: str = 'resnet50',
        backbone: str = 'resnet50',
        weights: WeightsEnum | str | bool | None = None,
        in_channels: int = 3,
        num_outputs: int = 1,
        num_filters: int = 3,
        loss: str = 'mse',
        lr: float = 1e-3,
        patience: int = 10,
        freeze_backbone: bool = False,
        freeze_decoder: bool = False,
    ) -> None:
        """Initialize a new RegressionTask instance.

        Args:
            model: Name of the
                `timm <https://huggingface.co/docs/timm/reference/models>`__ or
                `smp <https://smp.readthedocs.io/en/latest/models.html>`__ model to use.
            backbone: Name of the
                `timm <https://smp.readthedocs.io/en/latest/encoders_timm.html>`__ or
                `smp <https://smp.readthedocs.io/en/latest/encoders.html>`__ backbone
                to use. Only applicable to PixelwiseRegressionTask.
            weights: Initial model weights. Either a weight enum, the string
                representation of a weight enum, True for ImageNet weights, False
                or None for random weights, or the path to a saved model state dict.
            in_channels: Number of input channels to model.
            num_outputs: Number of prediction outputs.
            num_filters: Number of filters. Only applicable when model='fcn'.
            loss: One of 'mse' or 'mae'.
            lr: Learning rate for optimizer.
            patience: Patience for learning rate scheduler.
            freeze_backbone: Freeze the backbone network to linear probe
                the regression head. Does not support FCN models.
            freeze_decoder: Freeze the decoder network to linear probe
                the regression head. Does not support FCN models.
                Only applicable to PixelwiseRegressionTask.

        .. versionchanged:: 0.4
           Change regression model support from torchvision.models to timm

        .. versionadded:: 0.5
           The *freeze_backbone* and *freeze_decoder* parameters.

        .. versionchanged:: 0.5
           *learning_rate* and *learning_rate_schedule_patience* were renamed to
           *lr* and *patience*.
        """
        self.weights = weights
        super().__init__()

    def configure_models(self) -> None:
        """Initialize the model."""
        # Create model
        weights = self.weights
        self.model = timm.create_model(
            self.hparams['model'],
            num_classes=self.hparams['num_outputs'],
            in_chans=self.hparams['in_channels'],
            pretrained=weights is True,
        )

        # Load weights
        if weights and weights is not True:
            if isinstance(weights, WeightsEnum):
                state_dict = weights.get_state_dict(progress=True)
            elif os.path.exists(weights):
                _, state_dict = utils.extract_backbone(weights)
            else:
                state_dict = get_weight(weights).get_state_dict(progress=True)
            utils.load_state_dict(self.model, state_dict)

        # Freeze backbone and unfreeze classifier head
        if self.hparams['freeze_backbone']:
            for param in self.model.parameters():
                param.requires_grad = False
            for param in self.model.get_classifier().parameters():
                param.requires_grad = True

    def configure_losses(self) -> None:
        """Initialize the loss criterion.

        Raises:
            ValueError: If *loss* is invalid.
        """
        loss: str = self.hparams['loss']
        if loss == 'mse':
            self.criterion: nn.Module = nn.MSELoss()
        elif loss == 'mae':
            self.criterion = nn.L1Loss()
        else:
            raise ValueError(
                f"Loss type '{loss}' is not valid. "
                "Currently, supports 'mse' or 'mae' loss."
            )

    def configure_metrics(self) -> None:
        """Initialize the performance metrics.

        * :class:`~torchmetrics.MeanSquaredError`: The average of the squared
          differences between the predicted and actual values (MSE) and its
          square root (RMSE). Lower values are better.
        * :class:`~torchmetrics.MeanAbsoluteError`: The average of the absolute
          differences between the predicted and actual values (MAE).
          Lower values are better.
        """
        metrics = MetricCollection(
            {
                'RMSE': MeanSquaredError(squared=False),
                'MSE': MeanSquaredError(squared=True),
                'MAE': MeanAbsoluteError(),
            }
        )
        self.train_metrics = metrics.clone(prefix='train_')
        self.val_metrics = metrics.clone(prefix='val_')
        self.test_metrics = metrics.clone(prefix='test_')

    def training_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> Tensor:
        """Compute the training loss and additional metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.

        Returns:
            The loss tensor.
        """
        x = batch['image']
        batch_size = x.shape[0]
        # TODO: remove .to(...) once we have a real pixelwise regression dataset
        y = batch[self.target_key].to(torch.float)
        y_hat = self(x)
        if y_hat.ndim != y.ndim:
            y = y.unsqueeze(dim=1)
        loss: Tensor = self.criterion(y_hat, y)
        self.log('train_loss', loss, batch_size=batch_size)
        self.train_metrics(y_hat, y)
        self.log_dict(self.train_metrics, batch_size=batch_size)

        return loss

    def validation_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        """Compute the validation loss and additional metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """
        x = batch['image']
        batch_size = x.shape[0]
        # TODO: remove .to(...) once we have a real pixelwise regression dataset
        y = batch[self.target_key].to(torch.float)
        y_hat = self(x)
        if y_hat.ndim != y.ndim:
            y = y.unsqueeze(dim=1)
        loss = self.criterion(y_hat, y)
        self.log('val_loss', loss, batch_size=batch_size)
        self.val_metrics(y_hat, y)
        self.log_dict(self.val_metrics, batch_size=batch_size)

        if (
            batch_idx < 10
            and hasattr(self.trainer, 'datamodule')
            and hasattr(self.trainer.datamodule, 'plot')
            and self.logger
            and hasattr(self.logger, 'experiment')
            and hasattr(self.logger.experiment, 'add_figure')
        ):
            datamodule = self.trainer.datamodule
            aug = K.AugmentationSequential(
                K.Denormalize(datamodule.mean, datamodule.std),
                data_keys=None,
                keepdim=True,
            )
            batch = aug(batch)
            if self.target_key == 'mask':
                y = y.squeeze(dim=1)
                y_hat = y_hat.squeeze(dim=1)
            batch['prediction'] = y_hat
            for key in ['image', self.target_key, 'prediction']:
                batch[key] = batch[key].cpu()
            sample = unbind_samples(batch)[0]

            fig: Figure | None = None
            try:
                fig = datamodule.plot(sample)
            except RGBBandsMissingError:
                pass

            if fig:
                summary_writer = self.logger.experiment
                summary_writer.add_figure(
                    f'image/{batch_idx}', fig, global_step=self.global_step
                )
                plt.close()

    def test_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> None:
        """Compute the test loss and additional metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """
        x = batch['image']
        batch_size = x.shape[0]
        # TODO: remove .to(...) once we have a real pixelwise regression dataset
        y = batch[self.target_key].to(torch.float)
        y_hat = self(x)
        if y_hat.ndim != y.ndim:
            y = y.unsqueeze(dim=1)
        loss = self.criterion(y_hat, y)
        self.log('test_loss', loss, batch_size=batch_size)
        self.test_metrics(y_hat, y)
        self.log_dict(self.test_metrics, batch_size=batch_size)

    def predict_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> Tensor:
        """Compute the predicted regression values.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.

        Returns:
            Output predicted probabilities.
        """
        x = batch['image']
        y_hat: Tensor = self(x)
        return y_hat


class PixelwiseRegressionTask(RegressionTask):
    """LightningModule for pixelwise regression of images.

    .. versionadded:: 0.5
    """

    target_key = 'mask'

    def configure_models(self) -> None:
        """Initialize the model."""
        weights = self.weights

        if self.hparams['model'] == 'unet':
            self.model = smp.Unet(
                encoder_name=self.hparams['backbone'],
                encoder_weights='imagenet' if weights is True else None,
                in_channels=self.hparams['in_channels'],
                classes=1,
            )
        elif self.hparams['model'] == 'deeplabv3+':
            self.model = smp.DeepLabV3Plus(
                encoder_name=self.hparams['backbone'],
                encoder_weights='imagenet' if weights is True else None,
                in_channels=self.hparams['in_channels'],
                classes=1,
            )
        elif self.hparams['model'] == 'fcn':
            self.model = FCN(
                in_channels=self.hparams['in_channels'],
                classes=1,
                num_filters=self.hparams['num_filters'],
            )
        else:
            raise ValueError(
                f"Model type '{self.hparams['model']}' is not valid. "
                "Currently, only supports 'unet', 'deeplabv3+' and 'fcn'."
            )

        if self.hparams['model'] != 'fcn':
            if weights and weights is not True:
                if isinstance(weights, WeightsEnum):
                    state_dict = weights.get_state_dict(progress=True)
                elif os.path.exists(weights):
                    _, state_dict = utils.extract_backbone(weights)
                else:
                    state_dict = get_weight(weights).get_state_dict(progress=True)
                self.model.encoder.load_state_dict(state_dict)

        # Freeze backbone
        if self.hparams.get('freeze_backbone', False) and self.hparams['model'] in [
            'unet',
            'deeplabv3+',
        ]:
            for param in self.model.encoder.parameters():
                param.requires_grad = False

        # Freeze decoder
        if self.hparams.get('freeze_decoder', False) and self.hparams['model'] in [
            'unet',
            'deeplabv3+',
        ]:
            for param in self.model.decoder.parameters():
                param.requires_grad = False



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""SpaceNet 6 dataset."""

from typing import ClassVar

from .base import SpaceNet


class SpaceNet6(SpaceNet):
    r"""SpaceNet 6: Multi-Sensor All-Weather Mapping.

    `SpaceNet 6 <https://spacenet.ai/sn6-challenge/>`_ is a dataset
    of optical and SAR imagery over the city of Rotterdam.

    Collection features:

    +------------+---------------------+------------+-----------------------------+
    |    AOI     | Area (km\ :sup:`2`\)| # Images   | # Building Footprint Labels |
    +============+=====================+============+=============================+
    | Rotterdam  |    120              |   3401     |         48000               |
    +------------+---------------------+------------+-----------------------------+


    Imagery features:

    .. list-table::
        :widths: 10 10 10 10 10 10
        :header-rows: 1
        :stub-columns: 1

        *   -
            - PAN
            - RGBNIR
            - PS-RGB
            - PS-RGBNIR
            - SAR-Intensity
        *   - GSD (m)
            - 0.5
            - 2.0
            - 0.5
            - 0.5
            - 0.5
        *   - Chip size (px)
            - 900 x 900
            - 450 x 450
            - 900 x 900
            - 900 x 900
            - 900 x 900


    Dataset format:

    * Imagery - GeoTIFFs from Worldview-2 (optical) and Capella Space (SAR)

        * PAN.tif (Panchromatic)
        * RGBNIR.tif (Multispectral)
        * PS-RGB (Pansharpened RGB)
        * PS-RGBNIR (Pansharpened RGBNIR)
        * SAR-Intensity (SAR Intensity)

    * Labels - GeoJSON

        * labels.geojson

    If you use this dataset in your research, please cite the following paper:

    * https://arxiv.org/abs/2004.06500

    .. versionadded:: 0.4
    """

    file_regex = r'_tile_(\d+)\.'
    dataset_id = 'SN6_buildings'
    tarballs: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {11: ['SN6_buildings_AOI_11_Rotterdam_train.tar.gz']},
        'test': {11: ['SN6_buildings_AOI_11_Rotterdam_test_public.tar.gz']},
    }
    md5s: ClassVar[dict[str, dict[int, list[str]]]] = {
        'train': {11: ['10ca26d2287716e3b6ef0cf0ad9f946e']},
        'test': {11: ['a07823a5e536feeb8bb6b6f0cb43cf05']},
    }
    valid_aois: ClassVar[dict[str, list[int]]] = {'train': [11], 'test': [11]}
    valid_images: ClassVar[dict[str, list[str]]] = {
        'train': ['PAN', 'PS-RGB', 'PS-RGBNIR', 'RGBNIR', 'SAR-Intensity'],
        'test': ['SAR-Intensity'],
    }
    valid_masks = ('geojson_buildings',)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""SKy Images and Photovoltaic Power Dataset (SKIPP'D)."""

import os
from collections.abc import Callable
from typing import Any, ClassVar

import matplotlib.pyplot as plt
import numpy as np
import torch
from einops import rearrange
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_url, extract_archive, lazy_import


class SKIPPD(NonGeoDataset):
    """SKy Images and Photovoltaic Power Dataset (SKIPP'D).

    The `SKIPP'D dataset <https://purl.stanford.edu/dj417rh1007>`_
    contains ground-based fish-eye photos of the sky for solar
    forecasting tasks.

    Dataset Format:

    * .hdf5 file containing images and labels
    * .npy files with corresponding datetime timestamps

    Dataset Features:

    * fish-eye RGB images (64x64px)
    * power output measurements from 30-kW rooftop PV array
    * 1-min interval across 3 years (2017-2019)

    Nowcast task:

    * 349,372 images under the split key *trainval*
    * 14,003 images under the split key *test*

    Forecast task:

    * 130,412 images under the split key *trainval*
    * 2,462 images under the split key *test*
    * consists of a concatenated RGB time-series of 16
      time-steps

    If you use this dataset in your research, please cite:

    * https://doi.org/10.48550/arXiv.2207.00913

    .. note::

       This dataset requires the following additional library to be installed:

       * `<https://pypi.org/project/h5py/>`_ to load the dataset

    .. versionadded:: 0.5
    """

    url = 'https://hf.co/datasets/torchgeo/skippd/resolve/a16c7e200b4618cd93be3143cdb973e3f21498fa/{}'
    md5: ClassVar[dict[str, str]] = {
        'forecast': 'f4f3509ddcc83a55c433be9db2e51077',
        'nowcast': '0000761d403e45bb5f86c21d3c69aa80',
    }

    data_file_name = '2017_2019_images_pv_processed_{}.hdf5'
    zipfile_name = '2017_2019_images_pv_processed_{}.zip'

    valid_splits = ('trainval', 'test')

    valid_tasks = ('nowcast', 'forecast')

    dateformat = '%m/%d/%Y, %H:%M:%S'

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'trainval',
        task: str = 'nowcast',
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new Dataset instance.

        Args:
            root: root directory where dataset can be found
            split: one of "trainval", or "test"
            task: one of "nowcast", or "forecast"
            transforms: a function/transform that takes an input sample
                and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 after downloading files (may be slow)

        Raises:
            AssertionError: if ``task`` or ``split`` is invalid
            DatasetNotFoundError: If dataset is not found and *download* is False.
            DependencyNotFoundError: If h5py is not installed.
        """
        lazy_import('h5py')

        assert split in self.valid_splits, (
            f'Please choose one of these valid data splits {self.valid_splits}.'
        )
        self.split = split

        assert task in self.valid_tasks, (
            f'Please choose one of these valid tasks {self.valid_tasks}.'
        )
        self.task = task

        self.root = root
        self.transforms = transforms
        self.download = download
        self.checksum = checksum
        self._verify()

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        h5py = lazy_import('h5py')
        with h5py.File(
            os.path.join(self.root, self.data_file_name.format(self.task)), 'r'
        ) as f:
            num_datapoints: int = f[self.split]['pv_log'].shape[0]

        return num_datapoints

    def __getitem__(self, index: int) -> dict[str, str | Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        sample: dict[str, str | Tensor] = {'image': self._load_image(index)}
        sample.update(self._load_features(index))

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _load_image(self, index: int) -> Tensor:
        """Load the input image.

        Args:
            index: index of image to load

        Returns:
            image tensor at index
        """
        h5py = lazy_import('h5py')
        with h5py.File(
            os.path.join(self.root, self.data_file_name.format(self.task)), 'r'
        ) as f:
            arr = f[self.split]['images_log'][index]

        # forecast has dimension [16, 64, 64, 3] but reshape to [48, 64, 64]
        # https://github.com/yuhao-nie/Stanford-solar-forecasting-dataset/blob/main/models/SUNSET_forecast.ipynb
        if self.task == 'forecast':
            arr = rearrange(arr, 't h w c-> (t c) h w')
        else:
            arr = rearrange(arr, 'h w c -> c h w')

        tensor = torch.from_numpy(arr).to(torch.float32)
        return tensor

    def _load_features(self, index: int) -> dict[str, str | Tensor]:
        """Load label.

        Args:
            index: index of label to load

        Returns:
            label tensor at index
        """
        h5py = lazy_import('h5py')
        with h5py.File(
            os.path.join(self.root, self.data_file_name.format(self.task)), 'r'
        ) as f:
            label = f[self.split]['pv_log'][index]

        path = os.path.join(self.root, f'times_{self.split}_{self.task}.npy')
        datestring = np.load(path, allow_pickle=True)[index].strftime(self.dateformat)

        features: dict[str, str | Tensor] = {
            'label': torch.tensor(label, dtype=torch.float32),
            'date': datestring,
        }
        return features

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted files already exist
        pathname = os.path.join(self.root, self.data_file_name.format(self.task))
        if os.path.exists(pathname):
            return

        # Check if the zip files have already been downloaded
        pathname = os.path.join(self.root, self.zipfile_name.format(self.task))
        if os.path.exists(pathname):
            self._extract()
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()
        self._extract()

    def _download(self) -> None:
        """Download the dataset and extract it."""
        download_url(
            self.url.format(self.zipfile_name.format(self.task)),
            self.root,
            filename=self.zipfile_name.format(self.task),
            md5=self.md5[self.task] if self.checksum else None,
        )
        self._extract()

    def _extract(self) -> None:
        """Extract the dataset."""
        zipfile_path = os.path.join(self.root, self.zipfile_name.format(self.task))
        extract_archive(zipfile_path, self.root)

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        In the ``forecast`` task the latest image is plotted.

        Args:
            sample: a sample return by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample
        """
        if self.task == 'nowcast':
            image, label = sample['image'].permute(1, 2, 0), sample['label'].item()
        else:
            image, label = (
                sample['image'].permute(1, 2, 0).reshape(64, 64, 3, 16)[:, :, :, -1],
                sample['label'][-1].item(),
            )

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            prediction = sample['prediction'].item()

        fig, ax = plt.subplots(1, 1, figsize=(10, 10))

        ax.imshow(image / 255)
        ax.axis('off')

        if show_titles:
            title = f'Label: {label:.3f}'
            if showing_predictions:
                title += f'\nPrediction: {prediction:.3f}'
            ax.set_title(title)

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Digital Typhoon dataset."""

import glob
import os
import tarfile
from collections.abc import Callable, Sequence
from typing import Any, ClassVar, TypedDict

import matplotlib.pyplot as plt
import pandas as pd
import torch
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_url, lazy_import, percentile_normalization


class _SampleSequenceDict(TypedDict):
    """Sample sequence dictionary."""

    id: str
    seq_id: list[int]


class DigitalTyphoon(NonGeoDataset):
    """Digital Typhoon Dataset for Analysis Task.

    This dataset contains typhoon-centered images, derived from hourly infrared channel
    images captured by meteorological satellites. It incorporates data from multiple
    generations of the Himawari weather satellite, dating back to 1978. These images
    have been transformed into brightness temperatures and adjusted for varying
    satellite sensor readings, yielding a consistent spatio-temporal dataset that
    covers over four decades.

    See `the Digital Typhoon website
    <http://agora.ex.nii.ac.jp/digital-typhoon/dataset/>`_
    for more information about the dataset.

    Dataset features:

    * infrared channel images from the Himawari weather satellite (512x512 px)
      at 5km spatial resolution
    * auxiliary features such as wind speed, pressure, and more that can be used
      for regression or classification tasks
    * 1,099 typhoons and 189,364 images

    Dataset format:

    * hdf5 files containing the infrared channel images
    * .csv files containing the metadata for each image

    If you use this dataset in your research, please cite the following papers:

    * https://doi.org/10.20783/DIAS.664

    .. versionadded:: 0.6
    """

    valid_tasks = ('classification', 'regression')
    aux_file_name = 'aux_data.csv'

    valid_features = (
        'year',
        'month',
        'day',
        'hour',
        'grade',
        'lat',
        'lng',
        'pressure',
        'wind',
        'dir50',
        'long50',
        'short50',
        'dir30',
        'long30',
        'short30',
        'landfall',
        'intp',
    )

    url = 'https://hf.co/datasets/torchgeo/digital_typhoon/resolve/cf2f9ef89168d31cb09e42993d35b068688fe0df/WP.tar.gz{0}'

    md5sums: ClassVar[dict[str, str]] = {
        'aa': '9e77a5f74783f7909dee0fb936053b17',
        'ab': '46aebdcba6e4e2df1619e4a3d7e556bb',
    }

    min_input_clamp = 170.0
    max_input_clamp = 300.0

    data_root = 'WP'

    def __init__(
        self,
        root: Path = 'data',
        task: str = 'regression',
        features: Sequence[str] = ['wind'],
        targets: Sequence[str] = ['wind'],
        sequence_length: int = 3,
        min_feature_value: dict[str, float] | None = None,
        max_feature_value: dict[str, float] | None = None,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new Digital Typhoon dataset instance.

        Args:
            root: root directory where dataset can be found
            task: whether to load 'regression' or 'classification' labels
            features: which auxiliary features to return
            targets: which auxiliary features to use as targets
            sequence_length: length of the sequence to return
            min_feature_value: minimum value for each feature
            max_feature_value: maximum value for each feature
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: If any arguments are invalid.
            DatasetNotFoundError: If dataset is not found and *download* is False.
            DependencyNotFoundError: If h5py is not installed.
        """
        lazy_import('h5py')
        self.root = root
        self.transforms = transforms
        self.download = download
        self.checksum = checksum
        self.sequence_length = sequence_length

        self.min_feature_value = min_feature_value
        self.max_feature_value = max_feature_value

        assert task in self.valid_tasks, (
            f'Please choose one of {self.valid_tasks}, you provided {task}.'
        )
        self.task = task

        assert set(features).issubset(set(self.valid_features))
        self.features = features

        assert set(targets).issubset(set(self.valid_features))
        self.targets = targets

        self._verify()

        self.aux_df = pd.read_csv(
            os.path.join(root, self.data_root, self.aux_file_name)
        )
        self.aux_df['datetime'] = pd.to_datetime(
            self.aux_df[['year', 'month', 'day', 'hour']]
        )

        self.aux_df = self.aux_df.sort_values(['year', 'month', 'day', 'hour'])
        self.aux_df['seq_id'] = self.aux_df.groupby(['id']).cumcount()

        self.aux_df.columns = [str(col) for col in self.aux_df.columns]

        # Compute the hour difference between consecutive images per typhoon id
        self.aux_df['hour_diff_consecutive'] = (
            self.aux_df.sort_values(['id', 'datetime'])
            .groupby('id')['datetime']
            .diff()
            .dt.total_seconds()
            / 3600
        )

        # Compute the hour difference between the first and second entry
        self.aux_df['hour_diff_to_next'] = (
            self.aux_df.groupby('id')['datetime']
            .shift(-1)
            .sub(self.aux_df['datetime'])
            .abs()
            .dt.total_seconds()
            / 3600
        )

        self.aux_df['hour_diff'] = self.aux_df['hour_diff_consecutive'].combine_first(
            self.aux_df['hour_diff_to_next']
        )
        self.aux_df.drop(
            ['hour_diff_consecutive', 'hour_diff_to_next'], axis=1, inplace=True
        )

        # 0 hour difference is for the last time step of each typhoon sequence and want
        # to keep only images that have max 1 hour difference
        self.aux_df = self.aux_df[self.aux_df['hour_diff'] <= 1]
        # Filter out all ids that only have less than sequence_length entries
        self.aux_df = self.aux_df.groupby('id').filter(
            lambda x: len(x) >= self.sequence_length
        )

        # Filter aux_df according to min_target_value
        if self.min_feature_value is not None:
            for feature, min_value in self.min_feature_value.items():
                self.aux_df = self.aux_df[self.aux_df[feature] >= min_value]

        # Filter aux_df according to max_target_value
        if self.max_feature_value is not None:
            for feature, max_value in self.max_feature_value.items():
                self.aux_df = self.aux_df[self.aux_df[feature] <= max_value]

        # collect target mean and std for each target
        self.target_mean: dict[str, float] = self.aux_df[self.targets].mean().to_dict()
        self.target_std: dict[str, float] = self.aux_df[self.targets].std().to_dict()

        def _get_subsequences(df: pd.DataFrame, k: int) -> list[dict[str, list[int]]]:
            """Generate all possible subsequences of length k for a given group.

            Args:
                df: grouped dataframe of a single typhoon
                k: length of the subsequences to generate

            Returns:
                list of all possible subsequences of length k for a given typhoon id
            """
            min_seq_id = df['seq_id'].min()
            max_seq_id = df['seq_id'].max()

            # generate possible subsquences of length k for group
            subsequences = [
                {'id': df['id'].iloc[0], 'seq_id': list(range(i, i + k))}
                for i in range(min_seq_id, max_seq_id - k + 2)
            ]
            return [
                subseq
                for subseq in subsequences
                if set(subseq['seq_id']).issubset(df['seq_id'])
            ]

        self.sample_sequences: list[_SampleSequenceDict] = [
            item
            for sublist in self.aux_df.groupby('id')[['seq_id', 'id']]
            .apply(_get_subsequences, k=self.sequence_length)
            .tolist()
            for item in sublist
        ]

    def __getitem__(self, index: int) -> dict[str, Any]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data, labels, and metadata at that index
        """
        sample_entry = self.sample_sequences[index]
        sample_df = self.aux_df[
            (self.aux_df['id'] == sample_entry['id'])
            & (self.aux_df['seq_id'].isin(sample_entry['seq_id']))
        ]

        sample = {'image': self._load_image(sample_df)}
        # load features of the last image in the sequence
        sample.update(
            self._load_features(
                os.path.join(
                    self.root,
                    self.data_root,
                    'metadata',
                    str(sample_df.iloc[-1]['id']) + '.csv',
                ),
                sample_df.iloc[-1]['image_path'],
            )
        )

        # torchgeo expects a single label
        sample['label'] = torch.Tensor([sample[target] for target in self.targets])

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.sample_sequences)

    def _load_image(self, sample_df: pd.DataFrame) -> Tensor:
        """Load a single image.

        Args:
            sample_df: df holding all information necessary to load the
                consecutive images in the sequence

        Returns:
            concatenation of all images in the sequence over channel dimension
        """

        def load_image_tensor(id: str, filepath: str) -> Tensor:
            """Load a single image tensor from a h5 file.

            Args:
                id: typhoon id
                filepath: path to the h5 file

            Returns:
                image tensor
            """
            h5py = lazy_import('h5py')

            full_path = os.path.join(self.root, self.data_root, 'image', id, filepath)
            with h5py.File(full_path, 'r') as h5f:
                # tensor with added channel dimension
                tensor = torch.from_numpy(h5f['Infrared'][:]).unsqueeze(0)

                # follow normalization procedure
                # https://github.com/kitamoto-lab/benchmarks/blob/1bdbefd7c570cb1bdbdf9e09f9b63f7c22bbdb27/analysis/regression/FrameDatamodule.py#L94
                tensor = torch.clamp(tensor, self.min_input_clamp, self.max_input_clamp)
                tensor = (tensor - self.min_input_clamp) / (
                    self.max_input_clamp - self.min_input_clamp
                )
            return tensor

        # tensor of shape [sequence_length, height, width]
        tensor = torch.cat(
            [
                load_image_tensor(str(id), filepath)
                for id, filepath in zip(sample_df['id'], sample_df['image_path'])
            ]
        ).float()
        return tensor

    def _load_features(self, filepath: str, image_path: str) -> dict[str, Any]:
        """Load features for the corresponding image.

        Args:
            filepath: path of the feature file to load
            image_path: image path for the unique image for which to retrieve features

        Returns:
            features for image
        """
        feature_df = pd.read_csv(filepath)
        feature_df = feature_df[feature_df['file_1'] == image_path]
        feature_dict = {
            name: torch.tensor(feature_df[name].item()).float()
            for name in self.features
        }
        # normalize the targets for regression
        if self.task == 'regression':
            for feature, mean in self.target_mean.items():
                feature_dict[feature] = (
                    feature_dict[feature] - mean
                ) / self.target_std[feature]
        return feature_dict

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted files already exist
        exists = []
        path = os.path.join(self.root, self.data_root, 'image', '*', '*.h5')
        if glob.glob(path):
            exists.append(True)
        else:
            exists.append(False)

        # check if aux.csv file exists
        exists.append(
            os.path.exists(os.path.join(self.root, self.data_root, self.aux_file_name))
        )
        if all(exists):
            return

        # Check if the tar.gz files have already been downloaded
        exists = []
        for suffix in self.md5sums.keys():
            path = os.path.join(self.root, f'{self.data_root}.tar.gz{suffix}')
            exists.append(os.path.exists(path))

        if all(exists):
            self._extract()
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download amd extract the dataset
        self._download()
        self._extract()

    def _download(self) -> None:
        """Download the dataset."""
        for suffix, md5 in self.md5sums.items():
            download_url(
                self.url.format(suffix), self.root, md5=md5 if self.checksum else None
            )

    def _extract(self) -> None:
        """Extract the dataset."""
        # Extract tarball
        for suffix in self.md5sums.keys():
            with tarfile.open(
                os.path.join(self.root, f'{self.data_root}.tar.gz{suffix}')
            ) as tar:
                tar.extractall(path=self.root)

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample return by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample
        """
        image, label = sample['image'], sample['label']

        image = percentile_normalization(image)

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            prediction = sample['prediction']

        fig, ax = plt.subplots(1, 1, figsize=(10, 10))

        ax.imshow(image.permute(1, 2, 0))
        ax.axis('off')

        if show_titles:
            title_dict = {
                label_name: label[idx].item()
                for idx, label_name in enumerate(self.targets)
            }
            title = f'Label: {title_dict}'
            if showing_predictions:
                title_dict = {
                    label_name: prediction[idx].item()
                    for idx, label_name in enumerate(self.targets)
                }
                title += f'\nPrediction: {title_dict}'
            ax.set_title(title)

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig
